#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from torchoutil.utils.collections import flat_dict_of_dict, list_dict_to_dict_list


class TestCollections(TestCase):
    def test_example_1(self) -> None:
        x = {
            "a": 1,
            "b": {
                "a": 2,
                "b": 10,
            },
        }
        expected = {"a": 1, "b.a": 2, "b.b": 10}

        output = flat_dict_of_dict(x)
        self.assertDictEqual(output, expected)

    def test_example_2(self) -> None:
        x = {"a": ["hello", "world"], "b": 3}
        expected = {"a.0": "hello", "a.1": "world", "b": 3}

        output = flat_dict_of_dict(x, flat_iterables=True)
        self.assertDictEqual(output, expected)

    def test_example_3(self) -> None:
        x = {"a": {"b": 1}, "a.b": 2}
        expected = {"a.b": 2}

        output = flat_dict_of_dict(x)
        assert output == expected

    def test_example_4(self) -> None:
        x = {"a": {"b": 1}, "a.b": 2}
        self.assertRaises(ValueError, flat_dict_of_dict, x, overwrite=False)


class TestListDictToDictList(TestCase):
    def test_example_1(self) -> None:
        lst = [{"a": 1, "b": 2}, {"a": 4, "b": 3, "c": 5}]
        output = list_dict_to_dict_list(lst, default_val=0)
        self.assertDictEqual(output, {"a": [1, 4], "b": [2, 3], "c": [0, 5]})


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, List, Union

import torch
from torch import Tensor
from torch.types import Number

from torchoutil.nn.functional.get import _DEVICE_CUDA_IF_AVAILABLE, get_device


def get_inverse_perm(indices: Tensor, dim: int = -1) -> Tensor:
    """Return inverse permutation indices.
    The output will be a tensor of shape (..., N).

    Args:
        indices: Original permutation indices as tensor of shape (..., N).
        dim: Dimension of indices. defaults to -1.
    """
    arange = torch.arange(
        indices.shape[dim],
        dtype=indices.dtype,
        device=indices.device,
    )
    arange = arange.expand(*indices.shape)
    indices_inv = torch.empty_like(indices)
    indices_inv = indices_inv.scatter(dim, indices, arange)
    return indices_inv


def randperm_diff(
    size: int,
    generator: Union[None, int, torch.Generator] = None,
    device: Union[str, torch.device, None] = _DEVICE_CUDA_IF_AVAILABLE,
) -> Tensor:
    """This function ensure that every value i cannot be the element at index i.
    The output will be a tensor of shape (size,).

    Args:
        size: The number of indices. Cannot be < 2.
        seed: The seed or torch.Generator used to generate permutation.
        device: The PyTorch device of the output indices tensor.

    Example 1
    ----------
        >>> torch.randperm(5)
        tensor([1, 4, 2, 5, 0])  # 2 is the element of index 2 !
        >>> randperm_diff(5)
        tensor([2, 0, 4, 1, 3])
    """
    if size < 2:
        raise ValueError(f"Invalid argument {size=} < 2 for randperm_diff.")

    device = get_device(device)
    if isinstance(generator, int):
        generator = torch.Generator().manual_seed(generator)

    perm_kws: Dict[str, Any] = dict(generator=generator, device=device)
    arange = torch.arange(size, device=device)
    perm = torch.randperm(size, **perm_kws)

    while perm.eq(arange).any():
        perm = torch.randperm(size, **perm_kws)
    return perm


def get_perm_indices(t1: Tensor, t2: Tensor) -> Tensor:
    """Find permutation between two vectors t1 and t2 which contains values from 0 to N-1.

    Example 1::
    -----------
        >>> t1 = torch.as_tensor([0, 1, 2, 4, 3, 6, 5, 7])
        >>> t2 = torch.as_tensor([0, 2, 1, 4, 3, 5, 6, 7])
        >>> indices = get_perm_indices(t1, t2)
        >>> (t1[indices] == t2).all().item()
        True
    """
    i1 = (t1[..., None, :] == t2[..., :, None]).int().argmax(dim=-2)
    return i1


def insert_at_indices(
    x: Tensor,
    indices: Union[Tensor, List, Number],
    values: Union[Number, Tensor],
) -> Tensor:
    """Insert value(s) in vector at specified indices.

    Example 1::
    -----------
        >>> x = torch.as_tensor([1, 1, 2, 2, 2, 3])
        >>> indices = torch.as_tensor([2, 5])
        >>> values = 4
        >>> insert_values(x, indices, values)
        tensor([1, 1, 4, 2, 2, 2, 4, 3])
    """
    if x.ndim != 1:
        raise ValueError(
            f"Invalid argument number of dims. (found {x.ndim=} but expected 1)"
        )

    device = x.device
    if isinstance(indices, (int, float, bool)):
        indices = torch.as_tensor([indices], device=device, dtype=torch.long)
    elif isinstance(indices, list):
        indices = torch.as_tensor(indices, device=device, dtype=torch.long)

    out = torch.empty((x.shape[0] + indices.shape[0]), dtype=x.dtype, device=device)
    indices = indices + torch.arange(
        indices.shape[0], device=indices.device, dtype=indices.dtype
    )
    out[indices] = values
    mask = torch.full((out.shape[0],), True, dtype=torch.bool)
    mask[indices] = False
    out[mask] = x
    return out


def remove_at_indices(
    x: Tensor,
    indices: Union[Tensor, List, Number],
) -> Tensor:
    """Remove value(s) in vector at specified indices."""
    if x.ndim != 1:
        raise ValueError(
            f"Invalid argument number of dims. (found {x.ndim=} but expected 1)"
        )

    device = x.device
    if isinstance(indices, (int, float, bool)):
        indices = torch.as_tensor([indices], device=device, dtype=torch.long)
    elif isinstance(indices, list):
        indices = torch.as_tensor(indices, device=device, dtype=torch.long)

    indices = indices + torch.arange(
        indices.shape[0], device=device, dtype=indices.dtype
    )
    mask = torch.full((x.shape[0],), True, dtype=torch.bool)
    mask[indices] = False
    out = x[mask]
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from unittest import TestCase

import torch

from torchoutil.nn.functional.get import get_device
from torchoutil.nn.functional.multilabel import (
    indices_to_multihot,
    indices_to_names,
    multihot_to_indices,
    multihot_to_names,
    names_to_indices,
    names_to_multihot,
    probs_to_indices,
    probs_to_multihot,
    probs_to_names,
)


class TestMultilabel(TestCase):
    def test_indices_to_multihot_1(self) -> None:
        indices = [[1, 2], [0], [], [3]]
        num_classes = 4
        expected_multihot = torch.as_tensor(
            [[0, 1, 1, 0], [1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1]], dtype=torch.bool
        )

        result = indices_to_multihot(indices, num_classes=num_classes)

        self.assertTrue(torch.equal(result, expected_multihot))

    def test_probs_to_indices_1(self) -> None:
        probs = torch.as_tensor([[1.0, 0.4, 0.1, 0.9]])
        expected_indices = [[0, 3]]

        result = probs_to_indices(probs, threshold=0.5)

        self.assertListEqual(result, expected_indices)

    def test_probs_to_multihot_1(self) -> None:
        probs = torch.as_tensor([[1.0, 0.4, 0.1, 0.9]])
        expected_multihot = torch.as_tensor([[True, False, False, True]])

        result = probs_to_multihot(probs, threshold=0.5)

        self.assertTrue(torch.equal(result, expected_multihot))

    def test_probs_to_names_1(self) -> None:
        probs = torch.as_tensor([[1.0, 0.4, 0.1, 0.9]])
        expected_names = [["0", "3"]]

        num_classes = probs.shape[-1]
        idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))
        result = probs_to_names(probs, threshold=0.5, idx_to_name=idx_to_name)

        self.assertListEqual(result, expected_names)

    def test_convert_multihot(self) -> None:
        num_samples = int(torch.randint(0, 20, ()).item())
        num_classes = int(torch.randint(1, 20, ()).item())
        threshold = torch.rand(())
        idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))

        probs_1 = torch.rand(num_samples, num_classes)
        multihot_1 = probs_1.ge(threshold)

        indices_1 = multihot_to_indices(multihot_1)
        names_1 = indices_to_names(indices_1, idx_to_name)
        multihot_2 = names_to_multihot(names_1, idx_to_name)

        names_2 = multihot_to_names(multihot_2, idx_to_name)
        indices_2 = names_to_indices(names_2, idx_to_name)

        self.assertTrue(torch.equal(multihot_1, multihot_2))
        self.assertListEqual(names_1, names_2)
        self.assertListEqual(indices_1, indices_2)

    def test_ints_to_multihots(self) -> None:
        device = get_device()
        ints = torch.as_tensor([[0, 1, 1]], device=device)
        num_classes = 5
        multihots = indices_to_multihot(ints, num_classes, dtype=torch.int)
        expected = torch.as_tensor([[1, 1, 0, 0, 0]], device=device)

        self.assertEqual(multihots.shape, expected.shape)
        self.assertTrue(multihots.eq(expected).all(), f"{multihots=}")

    def test_convert_and_reconvert(self) -> None:
        device = get_device()
        multihots = torch.as_tensor(
            [
                [1, 1, 1],
                [0, 0, 0],
                [1, 1, 0],
                [0, 0, 1],
            ],
            device=device,
        )
        expected_ints = [[0, 1, 2], [], [0, 1], [2]]

        num_classes = multihots.shape[1]
        ints = multihot_to_indices(multihots)
        new_multihots = indices_to_multihot(
            ints, num_classes, dtype=torch.int, device=device
        )

        self.assertListEqual(ints, expected_ints)
        self.assertEqual(multihots.shape, new_multihots.shape)
        self.assertTrue(multihots.eq(new_multihots).all())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torch import Tensor, nn

from torchoutil.nn.functional.repeat import repeat_interleave_nd
from torchoutil.utils.collections import dump_dict


class Repeat(nn.Module):
    def __init__(self, *repeats: int) -> None:
        super().__init__()
        self.repeats = repeats

    def forward(self, x: Tensor) -> Tensor:
        return x.repeat(self.repeats)

    def extra_repr(self) -> str:
        return dump_dict(dict(repeats=self.repeats))


class RepeatInterleave(nn.Module):
    def __init__(self, repeats: int, dim: int) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return x.repeat_interleave(self.repeats, self.dim)

    def extra_repr(self) -> str:
        return dump_dict(dict(repeats=self.repeats, dim=self.dim))


class RepeatInterleaveNd(nn.Module):
    def __init__(self, repeats: int, dim: int) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return repeat_interleave_nd(x, self.repeats, self.dim)

    def extra_repr(self) -> str:
        return dump_dict(dict(repeats=self.repeats, dim=self.dim))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .activation import *
from .crop import *
from .mask import *
from .multiclass import *
from .multilabel import *
from .numpy import *
from .pad import *
from .repeat import *
from .tensor import *
from .typed import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Callable,
    Generic,
    Optional,
    Protocol,
    Sized,
    TypeVar,
    Union,
    runtime_checkable,
)

from torch.utils.data.dataset import Dataset

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)


class EmptyDataset(Dataset[None]):
    """Dataset placeholder. Raises StopIteration if __getitem__ is called."""

    def __getitem__(self, index) -> None:
        raise StopIteration

    def __len__(self) -> int:
        return 0


@runtime_checkable
class SizedDatasetLike(Generic[T], Protocol):
    __getitem__: Callable[..., T]

    def __len__(self) -> int:
        ...


class TransformWrapper(Generic[T, U], Dataset[U]):
    def __init__(
        self,
        dataset: Dataset[T],
        transform: Optional[Callable[[T], U]],
    ) -> None:
        super().__init__()
        self._dataset = dataset
        self._transform = transform

    def __getitem__(self, index) -> Union[T, U]:
        item = self._dataset[index]
        if self._transform is not None:
            item = self._transform(item)
        return item

    def __len__(self) -> int:
        if isinstance(self._dataset, Sized):
            return len(self._dataset)
        else:
            raise TypeError("Wrapped dataset is not Sized.")

    def unwrap(self) -> Dataset[T]:
        return self._dataset

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torch import Tensor


def repeat_interleave_nd(x: Tensor, repeats: int, dim: int = 0) -> Tensor:
    """Generalized version of torch.repeat_interleave for N >= 1 dimensions.
    The output size will be (..., D*repeats, ...), where D is the size of the dimension of the dim argument.

    Args:
        x: Any tensor of shape (..., D, ...) with at least 1 dim.
        repeats: Number of repeats.
        dim: The dimension to repeat. defaults to 0.

    Examples::
    ----------
        >>> x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        >>> repeat_interleave_nd(x, n=2, dim=0)
        tensor([[0, 1, 2, 3],
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [4, 5, 6, 7]])
    """
    if x.ndim == 0:
        raise ValueError(
            f"Function repeat_interleave_nd does not supports 0-d tensors. (found {x.ndim=} == 0)"
        )

    dim = dim % x.ndim
    x = x.unsqueeze(dim=dim + 1)
    shape = list(x.shape)
    shape[dim + 1] = repeats
    x = x.expand(*shape)
    x = x.flatten(dim, dim + 1)
    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Collection of functions and modules to help development in PyTorch.
"""

__name__ = "torchoutil"
__author__ = "Étienne Labbé (Labbeti)"
__author_email__ = "labbeti.pub@gmail.com"
__license__ = "MIT"
__maintainer__ = "Étienne Labbé (Labbeti)"
__status__ = "Development"
__version__ = "0.2.2"


from .nn.functional import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import unittest
from unittest import TestCase

import numpy as np
import torch
from torch import Tensor
from torch.utils.data.dataset import Subset
from torchvision.datasets import CIFAR10

from torchoutil.nn import IndicesToOneHot, ToList, ToNumpy, TSequential
from torchoutil.utils.hdf import pack_to_hdf


class TestHDF(TestCase):
    def test_cifar10_pack_to_hdf(self) -> None:
        dataset = CIFAR10(
            "/tmp",
            train=False,
            transform=ToNumpy(),
            target_transform=TSequential(IndicesToOneHot(10), ToList()),
            download=True,
        )
        dataset = Subset(
            dataset,
            torch.randint(0, len(dataset), (max(len(dataset) // 10, 1),)).tolist(),
        )

        path = "/tmp/test_cifar10.hdf"
        hdf_dataset = pack_to_hdf(dataset, path, overwrite=True)

        idx = 0
        image0, label0 = dataset[idx]
        image1, label1 = hdf_dataset[idx]

        assert len(dataset) == len(hdf_dataset)
        assert np.equal(image0, image1).bool().all()
        assert label0 == label1

        hdf_dataset.close()
        os.remove(path)

    def test_shape_column(self) -> None:
        data = [torch.rand(10, 2), torch.rand(10, 5), torch.rand(10, 3)]
        data_shape = [(10, 1), (10, 2), (10, 1)]

        dataset = [
            {"data": data[i], "data_shape": data_shape[i]} for i in range(len(data))
        ]

        path = "/tmp/test_shape.hdf"
        hdf_dataset = pack_to_hdf(dataset, path, overwrite=True)

        assert len(hdf_dataset.added_columns) == 0

        for i, item in enumerate(iter(hdf_dataset)):
            assert set(item.keys()) == {"data", "data_shape"}

            data_i = item["data"]
            assert isinstance(data_i, Tensor)
            assert data_i.shape == data_shape[i]

        hdf_dataset.close()
        os.remove(path)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os


def get_auto_num_cpus() -> int:
    """Returns the number of CPUs available on Linux-based platforms.
    Useful for setting num_workers argument in DataLoaders.

    On Windows and MAC OS, this will just return the number of CPUs on this machine.
    """
    try:
        num_cpus = len(os.sched_getaffinity(0))
    except AttributeError:
        num_cpus = os.cpu_count()
        if num_cpus is None:
            num_cpus = 0
    return num_cpus



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Literal, Union

import torch
from torch import Generator, Tensor


CROP_ALIGNS = ("left", "right", "center", "random")
CropAlign = Literal["left", "right", "center", "random"]


def crop_dim(
    x: Tensor,
    target_length: int,
    align: CropAlign = "left",
    dim: int = -1,
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Generic function to crop a single dimension."""
    return crop_dims(x, [target_length], [align], [dim], generator)


def crop_dims(
    x: Tensor,
    target_lengths: Iterable[int],
    aligns: Union[  # type: ignore
        CropAlign,
        Iterable[CropAlign],
    ] = "left",
    dims: Union[Iterable[int], Literal["auto"]] = "auto",
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Generic function to crop multiple dimensions."""

    target_lengths = list(target_lengths)

    if isinstance(aligns, str):
        aligns = [aligns] * len(target_lengths)
    else:
        aligns = list(aligns)

    if dims == "auto":
        dims = list(range(-len(target_lengths), 0))
    else:
        dims = list(dims)

    if isinstance(generator, int):
        generator = Generator().manual_seed(generator)

    if len(target_lengths) != len(dims):
        raise ValueError(
            f"Invalid number of targets lengths ({len(target_lengths)}) with the number of dimensions ({len(dims)})."
        )

    if len(aligns) != len(dims):
        raise ValueError(
            f"Invalid number of aligns ({len(aligns)}) with the number of dimensions ({len(dims)})."
        )

    slices = [slice(None)] * len(x.shape)

    for target_length, dim, align in zip(target_lengths, dims, aligns):
        if x.shape[dim] <= target_length:
            continue

        if align == "left":
            start = 0
            end = target_length
        elif align == "right":
            start = x.shape[dim] - target_length
            end = None
        elif align == "center":
            diff = x.shape[dim] - target_length
            start = diff // 2 + diff % 2
            end = start + target_length
        elif align == "random":
            diff = x.shape[dim] - target_length
            start = torch.randint(low=0, high=diff, size=(), generator=generator).item()
            end = start + target_length
        else:
            raise ValueError(
                f"Invalid argument {align=}. (expected one of {CROP_ALIGNS})"
            )

        slices[dim] = slice(start, end)

    x = x[slices]
    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional, Union

import torch


_DEVICE_CUDA_IF_AVAILABLE = "cuda_if_available"


def get_device(
    device: Union[str, torch.device, None] = _DEVICE_CUDA_IF_AVAILABLE
) -> Optional[torch.device]:
    if device == _DEVICE_CUDA_IF_AVAILABLE:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    if isinstance(device, str):
        device = torch.device(device)
    return device



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from unittest import TestCase

import torch

from torch import Tensor

from torchoutil.nn.functional.indices import randperm_diff


class TestIndices(TestCase):
    def test_randperm_diff(self) -> None:
        for i in range(2, 10):
            perm = randperm_diff(i)
            arange = torch.arange(0, i, device=perm.device)

            self.assertIsInstance(perm, Tensor)
            self.assertEqual(perm.shape, arange.shape)
            self.assertFalse(perm.eq(arange).any(), f"{perm}; {arange}")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import List

from torch.optim.optimizer import Optimizer


def get_lr(optim: Optimizer, idx: int = 0) -> float:
    """
    Get the learning rate of the first group of an optimizer.

    Args:
        optim: The optimizer to get.
        idx: The group index of the learning rate in the optimizer. defaults to 0.
    """
    return get_lrs(optim)[idx]


def get_lrs(optim: Optimizer) -> List[float]:
    """
    Get the learning rates in all groups of an optimizer.

    Args:
        optim: The optimizer to get.
    """
    return [group["lr"] for group in optim.param_groups]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from torch import Generator, Tensor, nn

from torchoutil.nn.functional.crop import crop_dim, crop_dims, CropAlign
from torchoutil.utils.collections import dump_dict


class CropDim(nn.Module):
    def __init__(
        self,
        target_length: int,
        align: CropAlign = "left",
        dim: int = -1,
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.target_length = target_length
        self.align = align
        self.dim = dim
        self.generator = generator

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return crop_dim(
            x,
            self.target_length,
            self.align,  # type: ignore
            self.dim,
            self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_length=self.target_length,
                align=self.align,
                dim=self.dim,
            )
        )


class CropDims(nn.Module):
    def __init__(
        self,
        target_lengths: Iterable[int],
        aligns: Iterable[CropAlign] = ("left",),
        dims: Iterable[int] = (-1,),
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.target_lengths = target_lengths
        self.aligns = aligns
        self.dims = dims
        self.generator = generator

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return crop_dims(x, self.target_lengths, self.aligns, self.dims, self.generator)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_lengths=self.target_lengths,
                aligns=self.aligns,
                dims=self.dims,
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions for conversion between classes indices, onehot, names and probabilities for multiclass classification.
"""

from typing import List, Mapping, Sequence, TypeVar, Union

import torch
from torch import Tensor
from torch.nn import functional as F

from torchoutil.nn.functional.get import get_device

T = TypeVar("T")


def indices_to_onehot(
    indices: Union[Sequence[int], Tensor],
    num_classes: int,
    *,
    device: Union[str, torch.device, None] = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert indices of labels to onehot boolean encoding.

    Args:
        indices: List of list of label indices.
        num_classes: Number maximal of unique classes.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    device = get_device(device)
    indices = torch.as_tensor(indices, device=device, dtype=torch.long)
    onehot = F.one_hot(indices, num_classes)
    onehot = onehot.to(dtype=dtype)
    return onehot


def indices_to_names(
    indices: Union[Sequence[int], Tensor],
    idx_to_name: Mapping[int, T],
) -> List[T]:
    """Convert indices of labels to names using a mapping.

    Args:
        indices: List of list of label indices.
        idx_to_name: Mapping to convert a class index to its name.
    """
    return [idx_to_name[indices_i] for indices_i in indices]  # type: ignore


def onehot_to_indices(
    onehot: Tensor,
) -> List[int]:
    """Convert onehot boolean encoding to indices of labels.

    Args:
        onehot: OneHot labels encoded as 2D matrix.
    """
    return onehot.argmax(dim=-1).tolist()


def onehot_to_names(
    onehot: Tensor,
    idx_to_name: Mapping[int, T],
) -> List[T]:
    """Convert onehot boolean encoding to names using a mapping.

    Args:
        onehot: OneHot labels encoded as 2D matrix.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = onehot_to_indices(onehot)
    names = indices_to_names(indices, idx_to_name)
    return names


def names_to_indices(
    names: List[T],
    idx_to_name: Mapping[int, T],
) -> List[int]:
    """Convert names to indices of labels.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
    """
    name_to_idx = {name: idx for idx, name in idx_to_name.items()}
    indices = [name_to_idx[name] for name in names]
    return indices


def names_to_onehot(
    names: List[T],
    idx_to_name: Mapping[int, T],
    *,
    device: Union[str, torch.device, None] = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert names to onehot boolean encoding.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    indices = names_to_indices(names, idx_to_name)
    onehot = indices_to_onehot(indices, len(idx_to_name), device=device, dtype=dtype)
    return onehot


def probs_to_indices(
    probs: Tensor,
) -> List[int]:
    """Convert matrix of probabilities to indices of labels.

    Args:
        probs: Output probabilities for each classes.
    """
    indices = probs.argmax(dim=-1)
    indices = indices.tolist()
    return indices


def probs_to_onehot(
    probs: Tensor,
    *,
    device: Union[str, torch.device, None] = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert matrix of probabilities to onehot boolean encoding.

    Args:
        probs: Output probabilities for each classes.
    """
    if device is None:
        device = probs.device
    indices = probs_to_indices(probs)
    onehot = indices_to_onehot(indices, probs.shape[-1], device=device, dtype=dtype)
    return onehot


def probs_to_names(
    probs: Tensor,
    idx_to_name: Mapping[int, T],
) -> List[T]:
    """Convert matrix of probabilities to labels names.

    Args:
        probs: Output probabilities for each classes.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = probs_to_indices(probs)
    names = indices_to_names(indices, idx_to_name)
    return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions for conversion between classes indices, multihot, names and probabilities for multilabel classification.
"""

from typing import List, Mapping, Sequence, TypeVar, Union

import torch
from torch import Tensor

from torchoutil.nn.functional.get import get_device

T = TypeVar("T")


def indices_to_multihot(
    indices: Union[Sequence[Union[Sequence[int], Tensor]], Tensor],
    num_classes: int,
    *,
    device: Union[str, torch.device, None] = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert indices of labels to multihot boolean encoding.

    Args:
        indices: List of list of label indices.
        num_classes: Number maximal of unique classes.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    if device is None and isinstance(indices, Tensor):
        device = indices.device
    elif device is None and len(indices) > 0 and isinstance(indices[0], Tensor):
        device = indices[0].device  # type: ignore
    else:
        device = get_device(device)

    bsize = len(indices)
    multihot = torch.full((bsize, num_classes), False, dtype=dtype, device=device)
    for i, indices_i in enumerate(indices):
        if isinstance(indices_i, Tensor):
            indices_i = indices_i.to(device=device)
        else:
            indices_i = torch.as_tensor(indices_i, dtype=torch.long, device=device)
        multihot[i].scatter_(0, indices_i, True)
    return multihot


def indices_to_names(
    indices: Union[Sequence[Union[Sequence[int], Tensor]], Tensor],
    idx_to_name: Mapping[int, T],
) -> List[List[T]]:
    """Convert indices of labels to names using a mapping.

    Args:
        indices: List of list of label indices.
        idx_to_name: Mapping to convert a class index to its name.
    """
    names = []
    for indices_i in indices:
        names_i = [idx_to_name[idx] for idx in indices_i]  # type: ignore
        names.append(names_i)
    return names


def multihot_to_indices(
    multihot: Union[Tensor, Sequence[Tensor], Sequence[Sequence[bool]]],
) -> List[List[int]]:
    """Convert multihot boolean encoding to indices of labels.

    Args:
        multihot: Multihot labels encoded as 2D matrix.
    """
    preds = []
    for multihot_i in multihot:
        if not isinstance(multihot_i, Tensor):
            multihot_i = torch.as_tensor(multihot_i)
        preds_i = torch.where(multihot_i)[0].tolist()
        preds.append(preds_i)
    return preds


def multihot_to_names(
    multihot: Union[Tensor, Sequence[Tensor], Sequence[Sequence[bool]]],
    idx_to_name: Mapping[int, T],
) -> List[List[T]]:
    """Convert multihot boolean encoding to names using a mapping.

    Args:
        multihot: Multihot labels encoded as 2D matrix.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = multihot_to_indices(multihot)
    names = indices_to_names(indices, idx_to_name)
    return names


def names_to_indices(
    names: List[List[T]],
    idx_to_name: Mapping[int, T],
) -> List[List[int]]:
    """Convert names to indices of labels.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
    """
    name_to_idx = {name: idx for idx, name in idx_to_name.items()}
    indices = []
    for names_i in names:
        indices_i = [name_to_idx[name] for name in names_i]
        indices.append(indices_i)
    return indices


def names_to_multihot(
    names: List[List[T]],
    idx_to_name: Mapping[int, T],
    *,
    device: Union[str, torch.device, None] = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert names to multihot boolean encoding.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    indices = names_to_indices(names, idx_to_name)
    multihot = indices_to_multihot(
        indices,
        len(idx_to_name),
        device=device,
        dtype=dtype,
    )
    return multihot


def probs_to_indices(
    probs: Tensor,
    threshold: Union[float, Sequence[float], Tensor],
) -> List[List[int]]:
    """Convert matrix of probabilities to indices of labels.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
    """
    multihot = probs_to_multihot(probs, threshold)
    indices = multihot_to_indices(multihot)
    return indices


def probs_to_multihot(
    probs: Tensor,
    threshold: Union[float, Sequence[float], Tensor],
    *,
    device: Union[str, torch.device, None] = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert matrix of probabilities to multihot boolean encoding.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    if probs.ndim != 2:
        raise ValueError(
            "Invalid argument probs. (expected a batch of probabilities of shape (N, n_classes))."
        )

    num_classes = probs.shape[-1]
    if device is None:
        device = probs.device
    else:
        device = get_device(device)

    if not isinstance(threshold, Tensor):
        threshold = torch.as_tensor(threshold, dtype=torch.float, device=device)
    else:
        threshold = threshold.to(device=device)

    if threshold.ndim == 0:
        threshold = threshold.repeat(num_classes)
    elif threshold.ndim == 1:
        if threshold.shape[0] != num_classes:
            raise ValueError(
                f"Invalid argument threshold. (number of thresholds is {threshold.shape[0]} but found {num_classes} classes)"
            )
    else:
        raise ValueError(
            f"Invalid number of dimensions in input threshold. (found {threshold.shape=} but expected 0-d or 1-d tensor)"
        )

    multihot = probs >= threshold
    multihot = multihot.to(dtype=dtype)
    return multihot


def probs_to_names(
    probs: Tensor,
    threshold: Union[float, Sequence[float], Tensor],
    idx_to_name: Mapping[int, T],
) -> List[List[T]]:
    """Convert matrix of probabilities to labels names.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = probs_to_indices(probs, threshold)
    names = indices_to_names(indices, idx_to_name)
    return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Module versions of tensor functions that do not already exists in PyTorch."""

from typing import List, Optional, Union

import torch
from torch import Tensor, nn
from torch.nn import functional as F
from torch.types import Number

from torchoutil.nn.functional.get import get_device
from torchoutil.utils import return_types
from torchoutil.utils.collections import dump_dict


class Reshape(nn.Module):
    def __init__(self, *shape: int) -> None:
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        return x.reshape(self.shape)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                shape=self.shape,
            ),
        )


class Mean(nn.Module):
    def __init__(self, dim: Union[int, None], keepdim: bool = False) -> None:
        super().__init__()
        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Tensor:
        if self.dim is None:
            return x.mean()
        else:
            return x.mean(dim=self.dim, keepdim=self.keepdim)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
            ),
            ignore_none=True,
        )


class Max(nn.Module):
    def __init__(
        self,
        dim: Optional[int],
        return_values: bool = True,
        return_indices: Optional[bool] = None,
        keepdim: bool = False,
    ) -> None:
        if return_indices is None:
            return_indices = dim is not None
        if not return_values and not return_indices:
            raise ValueError(
                f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            )
        if dim is None and keepdim:
            raise ValueError(
                f"Invalid combinaison of arguments {dim=} and {keepdim=}. (expected dim is not None or keepdim=False)"
            )

        super().__init__()
        self.dim = dim
        self.return_values = return_values
        self.return_indices = return_indices
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Union[Tensor, return_types.max]:
        if self.dim is None:
            index = x.argmax()
            values_indices = return_types.max(x.flatten()[index], index)
        else:
            values_indices = x.max(dim=self.dim, keepdim=self.keepdim)

        if self.return_values and self.return_indices:
            return values_indices
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            raise ValueError(
                f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                return_values=self.return_values,
                return_indices=self.return_indices,
                keepdim=self.keepdim,
            ),
        )


class Min(nn.Module):
    def __init__(
        self,
        dim: Optional[int],
        return_values: bool = True,
        return_indices: Optional[bool] = None,
        keepdim: bool = False,
    ) -> None:
        if return_indices is None:
            return_indices = dim is not None
        if not return_values and not return_indices:
            raise ValueError(
                f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            )
        if dim is None and keepdim:
            raise ValueError(
                f"Invalid combinaison of arguments {dim=} and {keepdim=}. (expected dim is not None or keepdim=False)"
            )

        super().__init__()
        self.dim = dim
        self.return_values = return_values
        self.return_indices = return_indices
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Union[Tensor, return_types.min]:
        if self.dim is None:
            index = x.argmin()
            values_indices = return_types.min(x.flatten()[index], index)
        else:
            values_indices = x.min(dim=self.dim, keepdim=self.keepdim)

        if self.return_values and self.return_indices:
            return values_indices
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            raise ValueError(
                f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                return_values=self.return_values,
                return_indices=self.return_indices,
                keepdim=self.keepdim,
            ),
        )


class Normalize(nn.Module):
    def __init__(
        self,
        p: float = 2.0,
        dim: int = 1,
        eps: float = 1e-12,
    ) -> None:
        super().__init__()
        self.p = p
        self.dim = dim
        self.eps = eps

    def forward(self, x: Tensor) -> Tensor:
        return F.normalize(x, self.p, self.dim, self.eps)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                p=self.p,
                dim=self.dim,
                eps=self.eps,
            )
        )


class Transpose(nn.Module):
    def __init__(self, dim0: int, dim1: int) -> None:
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def forward(self, x: Tensor) -> Tensor:
        return x.transpose(self.dim0, self.dim1)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim0=self.dim0,
                dim1=self.dim1,
            ),
            fmt="{value}",
        )


class Squeeze(nn.Module):
    def __init__(self, dim: Optional[int] = None, inplace: bool = True) -> None:
        super().__init__()
        self.dim = dim
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if not self.inplace:
            if self.dim is None:
                return x.squeeze()
            else:
                return x.squeeze(self.dim)
        else:
            if self.dim is None:
                return x.squeeze_()
            else:
                return x.squeeze_(self.dim)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                inplace=self.inplace,
            ),
        )


class Unsqueeze(nn.Module):
    def __init__(self, dim: int, inplace: bool = True) -> None:
        super().__init__()
        self.dim = dim
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if not self.inplace:
            return x.unsqueeze(self.dim)
        else:
            return x.unsqueeze_(self.dim)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                inplace=self.inplace,
            ),
        )


class TensorTo(nn.Module):
    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.kwargs = kwargs

    def forward(self, x: Tensor) -> Tensor:
        return x.to(**self.kwargs)

    def extra_repr(self) -> str:
        return dump_dict(dict(self.kwargs))


class Permute(nn.Module):
    def __init__(self, *args: int) -> None:
        super().__init__()
        self.dims = tuple(args)

    def forward(self, x: Tensor) -> Tensor:
        return x.permute(self.dims)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dims=self.dims,
            ),
            fmt="{value}",
        )


class ToList(nn.Module):
    def forward(self, x: Tensor) -> List:
        return x.tolist()


class ToItem(nn.Module):
    def forward(self, x: Tensor) -> Number:
        return x.item()


class AsTensor(nn.Module):
    def __init__(
        self,
        device: Union[str, torch.device, None] = None,
        dtype: Union[torch.dtype, None] = None,
    ) -> None:
        device = get_device(device)
        super().__init__()
        self.device = device
        self.dtype = dtype

    def forward(self, x: Union[List, Number, Tensor]) -> Tensor:
        return torch.as_tensor(x, dtype=self.dtype, device=self.device)  # type: ignore

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_none=True,
        )


class Abs(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return x.abs()


class Angle(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return x.angle()


class Real(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return x.real


class Imag(nn.Module):
    def __init__(self, return_zeros: bool = False) -> None:
        """Return the imaginary part of a complex tensor.

        Args:
            return_zeros: If the input is not a complex tensor and return_zeros=True, the module will return a tensor containing zeros.
        """
        super().__init__()
        self.return_zeros = return_zeros

    def forward(self, x: Tensor) -> Tensor:
        if self.return_zeros and not x.is_complex():
            return torch.zeros_like(x)
        else:
            return x.imag



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import sys
from pathlib import Path
from typing import Dict

import torch

import torchoutil
from torchoutil.utils.collections import dump_dict


def get_package_repository_path() -> str:
    """Return the absolute path where the source code of this package is installed."""
    return str(Path(__file__).parent.parent.parent)


def get_install_info() -> Dict[str, str]:
    return {
        "torchoutil": torchoutil.__version__,
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "os": platform.system(),
        "architecture": platform.architecture()[0],
        "torch": str(torch.__version__),
        "package_path": get_package_repository_path(),
    }


def print_install_info() -> None:
    """Show main packages versions."""
    install_info = get_install_info()
    dumped = dump_dict(install_info, join="\n", fmt="{key}: {value}")
    print(dumped)


if __name__ == "__main__":
    print_install_info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from torch import Tensor, nn

from torchoutil.nn.functional.activation import softmax_multidim
from torchoutil.utils.collections import dump_dict


class SoftmaxMultidim(nn.Module):
    def __init__(
        self,
        dims: Union[Iterable[int], None] = (-1,),
    ) -> None:
        if dims is not None:
            dims = tuple(dims)

        super().__init__()
        self.dims = dims

    def forward(
        self,
        input: Tensor,
    ) -> Tensor:
        input = softmax_multidim(input, self.dims)
        return input

    def extra_repr(self) -> str:
        return dump_dict(dict(dims=self.dims))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import os
import os.path as osp
import pickle
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import h5py
import numpy as np
import torch
from h5py import Dataset as HDFRawDataset
from torch import Tensor
from torch.utils.data.dataset import Dataset
from typing_extensions import TypeGuard

from torchoutil.nn.functional.indices import get_inverse_perm
from torchoutil.utils.collections import all_eq
from torchoutil.utils.hdf.common import (
    HDF_ENCODING,
    HDF_STRING_DTYPE,
    HDF_VOID_DTYPE,
    SHAPE_SUFFIX,
    _dict_to_tuple,
)
from torchoutil.utils.type_checks import (
    is_iterable_bytes_list,
    is_iterable_int,
    is_iterable_str,
)

logger = logging.getLogger(__name__)


T = TypeVar("T")
U = TypeVar("U")


IndexType = Union[int, Iterable[int], Tensor, slice, None]
ColumnType = Union[str, Iterable[str], None]


def _is_index(index: Any) -> TypeGuard[IndexType]:
    return (
        isinstance(index, int)
        or is_iterable_int(index)
        or isinstance(index, slice)
        or index is None
        or (isinstance(index, Tensor) and not index.is_floating_point())
    )


def _is_column(column: Any) -> TypeGuard[ColumnType]:
    return isinstance(column, str) or is_iterable_str(column) or column is None


class HDFDataset(Generic[T, U], Dataset[U]):
    def __init__(
        self,
        hdf_fpath: Union[str, Path],
        transform: Optional[Callable[[T], U]] = None,
        keep_padding: Iterable[str] = (),
        return_added_columns: bool = False,
        open_hdf: bool = True,
    ) -> None:
        """HDFDataset to read an packed hdf file.

        In HDF, all tensors are padded internally then cropped on-the-fly.
        Several options allows to extract non-padded tensors or return the internal shape of each column.

        Args:
            hdf_fpath: The path to the HDF file.
            transforms: The transform to apply values. default to None.
            keep_padding: Keys to keep padding values. defaults to ().
            return_added_columns: Returns the columns added by pack_to_hdf(.) function.
            open_hdf: If True, open the HDF file at start. defaults to True.
        """
        hdf_fpath = Path(hdf_fpath).resolve()
        if not hdf_fpath.is_file():
            names = os.listdir(osp.dirname(hdf_fpath))
            names = [name for name in names if name.endswith(".hdf")]
            names = list(sorted(names))
            names_str = "\n - ".join(names)
            raise FileNotFoundError(
                f"Cannot find HDF file in path {hdf_fpath=}. Possible HDF files are:\n - {names_str}"
            )
        keep_padding = list(keep_padding)

        super().__init__()
        self._hdf_fpath = hdf_fpath
        self._transform = transform
        self._keep_padding = keep_padding
        self._return_added_columns = return_added_columns

        self._hdf_file: Any = None

        if open_hdf:
            self.open()

    # Properties
    @property
    def added_columns(self) -> List[str]:
        """Return the list of columns added by pack_to_hdf function."""
        return list(self._hdf_file.attrs.get("added_columns", []))

    @property
    def all_columns(self) -> List[str]:
        """The name of all columns of the dataset."""
        return list(self.get_hdf_keys())

    @property
    def attrs(self) -> Dict[str, Any]:
        return dict(self._hdf_file.attrs)

    @property
    def column_names(self) -> List[str]:
        """The name of each column of the dataset."""
        column_names = self.all_columns
        column_names = [
            name
            for name in column_names
            if self._return_added_columns or name not in self.added_columns
        ]
        return column_names

    @property
    def shape(self) -> Tuple[int, ...]:
        """The shape of the Clotho dataset."""
        return len(self), len(self.column_names)

    @property
    def shape_suffix(self) -> str:
        """Return the tensor shape suffix in column names."""
        return self._hdf_file.attrs.get("shape_suffix", SHAPE_SUFFIX)

    @property
    def info(self) -> Dict[str, Any]:
        """Return the global dataset info."""
        return json.loads(self._hdf_file.attrs.get("info", "{}"))

    @property
    def item_type(self) -> str:
        """Return the global dataset info."""
        return str(self._hdf_file.attrs.get("item_type", "dict"))

    @property
    def num_columns(self) -> int:
        return len(self.column_names)

    @property
    def num_rows(self) -> int:
        return len(self)

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform

    # Public methods
    @overload
    def at(self, index: int) -> T:
        ...

    @overload
    def at(self, index: Union[Iterable[int], slice, None], column: str) -> List:
        ...

    @overload
    def at(self, index: Union[Iterable[int], slice, None]) -> Dict[str, List]:
        ...

    @overload
    def at(
        self,
        index: Union[Iterable[int], slice, None],
        column: Union[Iterable[str], None],
    ) -> Dict[str, List]:
        ...

    @overload
    def at(self, index: Any, column: Any) -> Any:
        ...

    def at(
        self,
        index: IndexType = None,
        column: ColumnType = None,
        raw: bool = False,
    ) -> Any:
        if not self.is_open():
            raise RuntimeError(
                f"Cannot get_raw value with closed HDF file. ({self._hdf_file is not None=} and {bool(self._hdf_file)=})"
            )

        if index is None:
            index = slice(None)
        elif isinstance(index, Tensor):
            index = index.tolist()
        if column is None:
            column = self.column_names

        if is_iterable_str(column):
            return {column_i: self.at(index, column_i) for column_i in column}

        if column not in self.column_names:
            raise ValueError(
                f"Invalid argument {column=}. (expected one of {tuple(self.column_names)})"
            )

        if isinstance(index, slice):
            is_mult = True
        elif is_iterable_int(index):
            is_mult = True
        elif isinstance(index, int):
            if not (-len(self) <= index < len(self)):
                raise IndexError(
                    f"Invalid argument {index=}. (expected int in range [{-len(self)}, {len(self)-1}])"
                )
            is_mult = False
        else:
            raise TypeError(f"Invalid argument type {type(index)=}.")

        hdf_value = self._raw_at(index, column)
        if raw:
            return hdf_value

        if is_mult:
            hdf_values = hdf_value
        else:
            hdf_values = [hdf_value]
        del hdf_value

        shape_name = f"{column}{self.shape_suffix}"
        must_remove_padding = (
            shape_name in self._hdf_file.keys() and column not in self._keep_padding
        )
        hdf_ds: HDFRawDataset = self._hdf_file[column]
        hdf_dtype = hdf_ds.dtype

        if must_remove_padding:
            shapes = self._raw_at(index, shape_name)
            if not is_mult:
                shapes = [shapes]
            slices_lst = [
                tuple(slice(shape_i) for shape_i in shape) for shape in shapes
            ]
        else:
            slices_lst = [None] * int(hdf_ds.shape[0])

        outputs = []

        for hdf_value, slices in zip(hdf_values, slices_lst):
            # Remove the padding part
            if slices is not None:
                hdf_value = hdf_value[slices]

            # Decode all bytes to string
            if hdf_dtype == HDF_STRING_DTYPE:
                hdf_value = _decode_rec(hdf_value, HDF_ENCODING)
            # Convert numpy.array to torch.Tensor
            elif isinstance(hdf_value, np.ndarray):
                if hdf_dtype != HDF_VOID_DTYPE:
                    hdf_value = torch.from_numpy(hdf_value)
                else:
                    hdf_value = hdf_value.tolist()
            # Convert numpy scalars
            elif np.isscalar(hdf_value) and hasattr(hdf_value, "item"):
                hdf_value = hdf_value.item()  # type: ignore

            outputs.append(hdf_value)

        if not is_mult:
            outputs = outputs[0]
        return outputs

    def close(self) -> None:
        if not self.is_open():
            raise RuntimeError("Cannot close the HDF file twice.")
        self._hdf_file.close()
        self._hdf_file = None

    def get_attrs(self) -> Dict[str, Any]:
        return self._hdf_file.attrs

    def get_hdf_fpath(self) -> Path:
        return self._hdf_fpath

    def get_hdf_keys(self) -> Tuple[str, ...]:
        if self.is_open():
            return tuple(self._hdf_file.keys())
        else:
            raise RuntimeError("Cannot get keys from a closed HDF file.")

    def get_column_shape(self, column_name: str) -> Tuple[int, ...]:
        if not self.is_open():
            msg = f"Cannot get max_shape with a closed HDF file. ({self._hdf_file is not None=} and {bool(self._hdf_file)=})"
            raise RuntimeError(msg)
        return tuple(self._hdf_file[column_name].shape)

    def is_closed(self) -> bool:
        return not self.is_open()

    def is_open(self) -> bool:
        return self._hdf_file is not None and bool(self._hdf_file)

    def open(self) -> None:
        if self.is_open():
            raise RuntimeError("Cannot open the HDF file twice.")
        self._hdf_file = h5py.File(self._hdf_fpath, "r")
        self._sanity_check()

    # Magic methods
    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, HDFDataset) and pickle.dumps(self) == pickle.dumps(__o)

    def __exit__(self) -> None:
        if self.is_open():
            self.close()

    @overload
    def __getitem__(self, index: int) -> U:
        ...

    @overload
    def __getitem__(
        self,
        index: Union[Iterable[int], slice, None],
    ) -> Dict[str, list]:
        ...

    @overload
    def __getitem__(self, index: Any) -> Any:
        ...

    def __getitem__(
        self,
        index: Union[IndexType, Tuple[IndexType, ColumnType]],
    ) -> Any:
        if (
            isinstance(index, tuple)
            and len(index) == 2
            and _is_index(index[0])
            and _is_column(index[1])
        ):
            index, column = index
        else:
            column = None

        item = self.at(index, column)  # type: ignore

        if isinstance(index, int) and (
            column is None
            or (
                isinstance(column, Iterable)
                and not isinstance(column, str)
                and set(column) == set(self.column_names)
            )
        ):
            if self.item_type == "tuple":
                item = _dict_to_tuple(item)
            if self._transform is not None:
                item = self._transform(item)  # type: ignore
        return item

    def __getstate__(self) -> Dict[str, Any]:
        return {
            "hdf_fpath": self._hdf_fpath,
            "transform": self._transform,
            "keep_padding": self._keep_padding,
            "return_added_columns": self._return_added_columns,
            "is_open": self.is_open(),
        }

    def __hash__(self) -> int:
        hash_value = 0
        if self.is_open():
            hash_value += self._hdf_file.attrs["global_hash_value"]
        if self._transform is not None:
            hash_value += hash(self._transform)
        hash_value += sum(map(hash, self._keep_padding))
        return hash_value

    def __len__(self) -> int:
        if self.is_closed():
            msg = f"Cannot length of a closed HDF file. ({self._hdf_file is not None=} and {bool(self._hdf_file)=})"
            raise RuntimeError(msg)

        if "length" in self._hdf_file.attrs:
            return self._hdf_file.attrs["length"]
        elif len(self._hdf_file) > 0:
            dataset: HDFRawDataset = next(iter(self._hdf_file.values()))
            return len(dataset)
        else:
            return 0

    def __repr__(self) -> str:
        repr_hparams = {"file": osp.basename(self._hdf_fpath), "size": len(self)}
        repr_ = ", ".join(f"{k}={v}" for k, v in repr_hparams.items())
        return f"HDFDataset({repr_})"

    def __setstate__(self, data: Dict[str, Any]) -> None:
        is_init = hasattr(self, "_hdf_fpath") and hasattr(self, "_hdf_file")
        files_are_different = is_init and self._hdf_fpath != data["hdf_fpath"]
        is_open = is_init and self.is_open()

        if is_init and files_are_different and is_open:
            self.close()

        self._hdf_fpath = data["hdf_fpath"]
        self._transform = data["transform"]
        self._keep_padding = data["keep_padding"]
        self._return_added_columns = data["return_added_columns"]
        self._hdf_file = None

        if not is_init or (files_are_different and is_open):
            self.open()

    # Private methods
    def _sanity_check(self) -> None:
        lens = [dset.shape[0] for dset in self._hdf_file.values()]
        if not all_eq(lens) or lens[0] != len(self):
            logger.error(
                f"Incorrect length stored in HDF file. (found {lens=} and {len(self)=})"
            )

        if hasattr(self, "__orig_class__"):
            t_type = self.__orig_class__.__args__[0]  # type: ignore
            if t_type is not Any and (
                (issubclass(t_type, dict) and self.item_type != "dict")
                or (issubclass(t_type, tuple) and self.item_type != "tuple")
            ):
                raise TypeError(
                    f"Invalid HDFDataset typing. (found specified type '{t_type.__name__}' but the internal dataset contains type '{self.item_type}')"
                )

    def _raw_at(self, index: Union[int, Iterable[int], slice], column: str) -> Any:
        if isinstance(index, Iterable):
            sorted_idxs, local_idxs = torch.as_tensor(index).sort(dim=-1)
            sorted_idxs = sorted_idxs.numpy()
            hdf_value: Any = self._hdf_file[column][sorted_idxs]
            inv_local_idxs = get_inverse_perm(local_idxs)
            hdf_value = [hdf_value[local_idx] for local_idx in inv_local_idxs]
        else:
            hdf_value: Any = self._hdf_file[column][index]
        return hdf_value


def _decode_rec(value: Union[bytes, Iterable], encoding: str) -> Union[str, list]:
    """Decode bytes to str with the specified encoding. Works recursively on list of bytes, list of list of bytes, etc."""
    if isinstance(value, bytes):
        return value.decode(encoding=encoding)
    elif is_iterable_bytes_list(value):
        return [_decode_rec(elt, encoding) for elt in value]
    else:
        raise TypeError(
            f"Invalid argument type {type(value)}. (expected bytes or Iterable)"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from torchoutil.utils.type_checks import is_mapping_str_any

T = TypeVar("T")
U = TypeVar("U")


def list_dict_to_dict_list(
    lst: Sequence[Mapping[str, T]],
    default_val: U = None,
    error_on_missing_key: bool = False,
) -> Dict[str, List[Union[T, U]]]:
    """Convert a list of dicts to a dict of lists.

    Example 1
    ----------
    >>> lst = [{'a': 1, 'b': 2}, {'a': 4, 'b': 3, 'c': 5}]
    >>> output = list_dict_to_dict_list(lst, default_val=0)
    {'a': [1, 4], 'b': [2, 3], 'c': [0, 5]}
    """
    if len(lst) == 0:
        return {}

    if error_on_missing_key:
        keys = set(lst[0])
        for dic in lst:
            if keys != set(dic.keys()):
                raise ValueError(
                    f"Invalid dict keys for list_dict_to_dict_list. (found {keys} and {dic.keys()})"
                )

    keys = {}
    for dic in lst:
        keys.update(dict.fromkeys(dic.keys()))

    out = {
        key: [
            lst[i][key] if key in lst[i].keys() else default_val
            for i in range(len(lst))
        ]
        for key in keys
    }
    return out


def dump_dict(
    dic: Mapping[str, Any],
    /,
    join: str = ", ",
    fmt: str = "{key}={value}",
    ignore_none: bool = False,
) -> str:
    """Dump dictionary to string."""
    result = join.join(
        fmt.format(key=key, value=value)
        for key, value in dic.items()
        if not (value is None and ignore_none)
    )
    return result


def pass_filter(
    name: T,
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
) -> bool:
    """Returns True if name in include set and not in exclude set."""
    if include is not None and exclude is not None:
        return (name in include) and (name not in exclude)
    if include is not None:
        return name in include
    elif exclude is not None:
        return name not in exclude
    else:
        return True


def filter_iterable(
    it: Iterable[T],
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
) -> List[T]:
    return [item for item in it if pass_filter(item, include, exclude)]


def all_eq(it: Iterable[T], eq_fn: Optional[Callable[[T, T], bool]] = None) -> bool:
    """Returns true if all elements in inputs are equal."""
    it = list(it)
    first = it[0]
    if eq_fn is None:
        return all(first == elt for elt in it)
    else:
        return all(eq_fn(first, elt) for elt in it)


def flat_dict_of_dict(
    nested_dic: Mapping[str, Any],
    sep: str = ".",
    flat_iterables: bool = False,
    overwrite: bool = True,
) -> Dict[str, Any]:
    """Flat a nested dictionary.

    Example 1
    ---------
    ```
    >>> dic = {
    ...     "a": 1,
    ...     "b": {
    ...         "a": 2,
    ...         "b": 10,
    ...     },
    ... }
    >>> flat_dict(dic)
    ... {"a": 1, "b.a": 2, "b.b": 10}
    ```

    Example 2
    ---------
    ```
    >>> dic = {"a": ["hello", "world"], "b": 3}
    >>> flat_dict(dic, flat_iterables=True)
    ... {"a.0": "hello", "a.1": "world", "b": 3}
    ```

    Args:
        nested_dict: Nested mapping containing sub-mappings or iterables.
        sep: Separators between keys.
        flat_iterables: If True, flat iterable and use index as key.
        overwrite: If True, overwrite duplicated keys in output. Otherwise duplicated keys will raises a ValueError.
    """
    output = {}
    for k, v in nested_dic.items():
        if is_mapping_str_any(v):
            v = flat_dict_of_dict(v, sep, flat_iterables)
            v = {f"{k}{sep}{kv}": vv for kv, vv in v.items()}
            output.update(v)

        elif flat_iterables and isinstance(v, Iterable) and not isinstance(v, str):
            v = {f"{i}": vi for i, vi in enumerate(v)}
            v = flat_dict_of_dict(v, sep, flat_iterables)
            v = {f"{k}{sep}{kv}": vv for kv, vv in v.items()}
            output.update(v)

        elif overwrite or k not in output:
            output[k] = v

        else:
            raise ValueError(f"Ambiguous flatten dict with key '{k}'.")

    return output


def flat_list(lst: Iterable[Sequence[T]]) -> Tuple[List[T], List[int]]:
    """Return a flat version of the input list of sublists with each sublist size."""
    flatten_lst = [element for sublst in lst for element in sublst]
    sizes = [len(sents) for sents in lst]
    return flatten_lst, sizes



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import os
import os.path as osp
from pathlib import Path
from typing import Dict, List, Optional, Tuple, TypedDict, Union

import torch
from torch import Tensor

from torchoutil.nn.functional.get import get_device

pylog = logging.getLogger(__name__)


class CheckpointInfo(TypedDict):
    architecture: str
    url: str
    hash: str
    fname: str


class ModelCheckpointRegister:
    def __init__(
        self,
        infos: Dict[str, CheckpointInfo],
        state_dict_key: Optional[str],
        ckpt_parent_path: Union[str, Path, None] = None,
    ) -> None:
        """
        Args:
            infos: Mapping model_name to their checkpoint information, with architecture, download url, hash value and filename.
            state_dict_key: Optional key in the checkpoint data where state_dict is stored.
            ckpt_parent_path: Directory where checkpoints are saved. If None, defaults to `~/.cache/torch/hub/checkpoints`.
        """
        if ckpt_parent_path is None:
            ckpt_parent_path = _default_ckpt_parent_path()
        else:
            ckpt_parent_path = Path(ckpt_parent_path)

        super().__init__()
        self._infos = infos
        self._state_dict_key = state_dict_key
        self._ckpt_parent_path = ckpt_parent_path

    @property
    def infos(self) -> Dict[str, CheckpointInfo]:
        return self._infos

    @property
    def state_dict_key(self) -> Optional[str]:
        return self._state_dict_key

    @property
    def ckpt_parent_path(self) -> Path:
        return self._ckpt_parent_path.resolve()

    @property
    def model_names(self) -> List[str]:
        return list(self._infos.keys())

    def get_ckpt_path(self, model_name: str) -> Path:
        if model_name not in self.model_names:
            raise ValueError(
                f"Invalid argument {model_name=}. (expected one of {self.model_names})"
            )

        fname = self._infos[model_name]["fname"]
        fpath = self.ckpt_parent_path.joinpath(fname)
        return fpath

    def load_state_dict(
        self,
        model_name_or_path: Union[str, Path],
        device: Union[str, torch.device, None] = None,
        offline: bool = False,
        verbose: int = 0,
    ) -> Dict[str, Tensor]:
        """Load state_dict weights.

        Args:
            model_name_or_path: Model name (case sensitive) or path to checkpoint file.
            device: Device of checkpoint weights.
            offline: If False, the checkpoint from a model name will be automatically downloaded.
            verbose: Verbose level. defaults to 0.

        Returns:
            State dict of model weights.
        """
        device = get_device(device)
        model_name_or_path = str(model_name_or_path)

        if osp.isfile(model_name_or_path):
            model_path = model_name_or_path
        else:
            try:
                model_path = self.get_ckpt_path(model_name_or_path)
            except ValueError:
                raise ValueError(
                    f"Invalid argument {model_name_or_path=}. (expected a path to a checkpoint file or a model name in {self.model_names})"
                )

            if not osp.isfile(model_path):
                if offline:
                    raise FileNotFoundError(
                        f"Cannot find checkpoint model file in '{model_path}' with mode {offline=}."
                    )
                else:
                    self.download_ckpt(model_name_or_path, verbose=verbose)

        del model_name_or_path

        data = torch.load(model_path, map_location=device)

        if self._state_dict_key is None:
            state_dict = data
        else:
            state_dict = data[self._state_dict_key]

        if verbose >= 1:
            test_map = data.get("test_mAP", "unknown")
            pylog.info(
                f"Loading encoder weights from '{model_path}'... (with test_mAP={test_map})"
            )

        return state_dict

    def download_ckpt(
        self,
        model_name: str,
        force: bool = False,
        verbose: int = 0,
    ) -> Tuple[Path, bool]:
        """Download checkpoint file."""
        fpath = self.get_ckpt_path(model_name)
        exists = fpath.exists()

        if exists and not force:
            return fpath, False

        if exists and force:
            os.remove(fpath)

        os.makedirs(fpath.parent, exist_ok=True)

        url = self._infos[model_name]["url"]
        torch.hub.download_url_to_file(url, str(fpath), progress=verbose >= 1)

        return fpath, True

    def save(self, path: Union[str, Path]) -> None:
        """Save info to JSON file."""
        args = {
            "infos": self._infos,
            "state_dict_key": self._state_dict_key,
            "ckpt_parent_path": str(self._ckpt_parent_path),
        }
        with open(path, "r") as file:
            json.dump(args, file)

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> "ModelCheckpointRegister":
        """Load register info from JSON file."""
        with open(path, "r") as file:
            args = json.load(file)
        return ModelCheckpointRegister(**args)


def _default_ckpt_parent_path() -> Path:
    """Default checkpoint path: `~/.cache/torch/hub/checkpoints`."""
    path = torch.hub.get_dir()
    path = Path(path)
    path = path.joinpath("checkpoints")
    return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import json
import logging
import zlib
from dataclasses import asdict, is_dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import h5py
import numpy as np
import torch
import tqdm
from h5py import Dataset as HDFRawDataset
from torch import Tensor, nn
from torch.utils.data.dataloader import DataLoader

from torchoutil.utils.collections import all_eq
from torchoutil.utils.data.dataloader import get_auto_num_cpus
from torchoutil.utils.data.dataset import SizedDatasetLike
from torchoutil.utils.hdf.common import (
    HDF_ENCODING,
    HDF_STRING_DTYPE,
    HDF_VOID_DTYPE,
    SHAPE_SUFFIX,
    _tuple_to_dict,
)
from torchoutil.utils.hdf.dataset import HDFDataset
from torchoutil.utils.type_checks import is_dict_str

pylog = logging.getLogger(__name__)

T = TypeVar("T")
U = TypeVar("U", bound=Union[int, float, str, Tensor, list])

_HDF_SUPPORTED_DTYPES = (
    "b",
    "i",
    "u",
    "f",
    HDF_STRING_DTYPE,
    HDF_VOID_DTYPE,
)
_HDFDType = Union[Literal["b", "i", "u", "f"], Any]


@torch.inference_mode()
def pack_to_hdf(
    dataset: SizedDatasetLike[T],
    hdf_fpath: Union[str, Path],
    pre_transform: Optional[Callable[[T], U]] = None,
    overwrite: bool = False,
    metadata: str = "",
    verbose: int = 0,
    batch_size: int = 32,
    num_workers: Union[int, Literal["auto"]] = "auto",
    shape_suffix: str = SHAPE_SUFFIX,
) -> HDFDataset[U, U]:
    """Pack a dataset to HDF file.

    Args:
        dataset: The sized dataset to pack. Must be sized and all items must be of dict type.
            The key of each dictionaries are strings and values can be int, float, str, Tensor, non-empty List[int], non-empty List[float], non-empty List[str].
            If values are tensors or lists, the number of dimensions must be the same for all items in the dataset.
        hdf_fpath: The path to the HDF file.
        pre_transform: The optional transform to apply to audio returned by the dataset BEFORE storing it in HDF file.
            Can be used for deterministic transforms like Resample, LogMelSpectrogram, etc. defaults to None.
        overwrite: If True, the file hdf_fpath can be overwritten. defaults to False.
        metadata: Additional metadata string to add to the hdf file. defaults to ''.
        verbose: Verbose level. defaults to 0.
        batch_size: The batch size of the dataloader. defaults to 32.
        num_workers: The number of workers of the dataloader.
            If "auto", it will be set to `len(os.sched_getaffinity(0))`. defaults to "auto".
    """
    # Check inputs
    if not isinstance(dataset, SizedDatasetLike):
        raise TypeError(
            f"Cannot pack to hdf a non-sized-dataset '{dataset.__class__.__name__}'."
        )
    if len(dataset) == 0:
        raise ValueError("Cannot pack to hdf an empty dataset.")

    hdf_fpath = Path(hdf_fpath).resolve()
    if hdf_fpath.exists() and not hdf_fpath.is_file():
        raise RuntimeError(f"Item {hdf_fpath=} exists but it is not a file.")

    if hdf_fpath.is_file() and not overwrite:
        raise ValueError(
            f"Cannot overwrite file {hdf_fpath}. Please remove it or use overwrite=True option."
        )

    if num_workers == "auto":
        num_workers = get_auto_num_cpus()
        if verbose >= 2:
            pylog.debug(f"Found num_workers=='auto', set to {num_workers}.")

    if pre_transform is None:
        pre_transform = nn.Identity()

    if verbose >= 2:
        pylog.debug(f"Start packing data into HDF file '{hdf_fpath}'...")

    # Step 1: Init max_shapes and hdf_dtypes with the first item
    shapes_0 = {}
    hdf_dtypes_0 = {}
    item_0 = dataset[0]
    item_0 = pre_transform(item_0)

    dict_pre_transform: Callable[[T], Dict[str, Any]]

    if is_dict_str(item_0):
        item_type = "dict"
        dict_pre_transform = pre_transform  # type: ignore
        item_0_dict = item_0
    elif isinstance(item_0, tuple):
        item_type = "tuple"
        dict_pre_transform = Compose(pre_transform, _tuple_to_dict)
        item_0_dict = _tuple_to_dict(item_0)
    else:
        raise ValueError(
            f"Invalid item type for {dataset.__class__.__name__}. (expected dict[str, Any] or tuple but found {type(item_0)})"
        )
    del pre_transform, item_0

    for attr_name, value in item_0_dict.items():
        shape, hdf_dtype = _get_shape_and_dtype(value)
        shapes_0[attr_name] = shape
        hdf_dtypes_0[attr_name] = hdf_dtype

    max_shapes: Dict[str, Tuple[int, ...]] = shapes_0
    hdf_dtypes: Dict[str, str] = hdf_dtypes_0
    all_eq_shapes: Dict[str, bool] = {
        attr_name: True for attr_name in item_0_dict.keys()
    }

    loader = DataLoader(
        dataset,  # type: ignore
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=nn.Identity(),
        drop_last=False,
        pin_memory=False,
    )

    for batch in tqdm.tqdm(
        loader,
        desc="Pre compute shapes...",
        disable=verbose <= 0,
    ):
        batch = [dict_pre_transform(item) for item in batch]

        for item in batch:
            for attr_name, value in item.items():
                shape, hdf_dtype = _get_shape_and_dtype(value)
                all_eq_shapes[attr_name] &= max_shapes[attr_name] == shape
                max_shapes[attr_name] = tuple(
                    map(max, zip(max_shapes[attr_name], shape))
                )
                if hdf_dtypes[attr_name] == hdf_dtype or hdf_dtype == HDF_VOID_DTYPE:
                    # Note: HDF_VOID_DTYPE is compatible
                    pass
                elif hdf_dtypes[attr_name] == HDF_VOID_DTYPE:
                    # Note: if the element 0 was void dtype, override with more specific dtype
                    hdf_dtypes[attr_name] = hdf_dtype
                else:
                    raise ValueError(
                        f"Found different hdf_dtype. (with {hdf_dtypes[attr_name]=} != {hdf_dtype=} and {attr_name=} with {value=})"
                    )

    if verbose >= 2:
        pylog.debug(f"Found max_shapes:\n{max_shapes}")
        pylog.debug(f"Found hdf_dtypes:\n{hdf_dtypes}")
        pylog.debug(f"Found all_eq_shapes:\n{all_eq_shapes}")

    now = datetime.datetime.now()
    creation_date = now.strftime("%Y-%m-%d_%H-%M-%S")

    if hasattr(dataset, "info"):
        info = dataset.info  # type: ignore
        if is_dataclass(info):
            info = asdict(info)
        elif isinstance(info, Mapping):
            info = dict(info.items())  # type: ignore
        else:
            info = {}
    else:
        info = {}

    added_columns = []

    with h5py.File(hdf_fpath, "w") as hdf_file:
        # Step 2: Build hdf datasets in file
        hdf_dsets: Dict[str, HDFRawDataset] = {}

        for attr_name, shape in max_shapes.items():
            hdf_dtype = hdf_dtypes.get(attr_name)

            kwargs: Dict[str, Any] = {}
            if hdf_dtype == "b":
                kwargs["fillvalue"] = False
            elif hdf_dtype in ("i", "u"):
                kwargs["fillvalue"] = 0
            elif hdf_dtype == "f":
                kwargs["fillvalue"] = 0.0
            elif hdf_dtype in (HDF_STRING_DTYPE, HDF_VOID_DTYPE):
                pass
            else:
                raise ValueError(
                    f"Unsupported type {hdf_dtype=}. (expected one of {_HDF_SUPPORTED_DTYPES})"
                )

            if verbose >= 2:
                pylog.debug(
                    f"Build hdf dset '{attr_name}' with shape={(len(dataset),) + shape}."
                )

            hdf_dsets[attr_name] = hdf_file.create_dataset(
                attr_name,
                (len(dataset),) + shape,
                hdf_dtype,
                **kwargs,
            )

        for attr_name, shape in max_shapes.items():
            if len(shape) == 0 or all_eq_shapes[attr_name]:
                continue

            shape_name = f"{attr_name}{shape_suffix}"
            raw_dset_shape = (len(dataset), len(shape))

            if shape_name in hdf_dsets:
                if hdf_dsets[shape_name].shape != raw_dset_shape:
                    msg = f"Column {shape_name} already exists in source dataset with a different shape. (found shape={hdf_dsets[shape_name].shape} but expected shape is {raw_dset_shape})"
                    raise RuntimeError(msg)
                else:
                    continue

            added_columns.append(shape_name)
            hdf_dsets[shape_name] = hdf_file.create_dataset(
                shape_name, raw_dset_shape, "i", fillvalue=-1
            )

        # Fill hdf datasets with a second pass through the whole dataset
        i = 0
        global_hash_value = 0

        loader = DataLoader(
            dataset,  # type: ignore
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            collate_fn=nn.Identity(),
            drop_last=False,
            pin_memory=False,
        )

        for batch in tqdm.tqdm(
            loader,
            desc="Pack data into HDF...",
            disable=verbose <= 0,
        ):
            batch = [dict_pre_transform(item) for item in batch]

            for item in batch:
                for attr_name, value in item.items():
                    hdf_dset = hdf_dsets[attr_name]
                    shape, hdf_dtype = _get_shape_and_dtype(value)

                    # Check every shape
                    if len(shape) != hdf_dset.ndim - 1:
                        raise ValueError(
                            f"Invalid number of dimension in audio (expected {len(shape)}, found {len(shape)})."
                        )

                    # Check dataset size
                    if any(
                        shape_i > dset_shape_i
                        for shape_i, dset_shape_i in zip(shape, hdf_dset.shape[1:])
                    ):
                        pylog.error(
                            f"Resize hdf_dset {attr_name} of shape {tuple(hdf_dset.shape[1:])} with new {shape=}."
                        )
                        raise RuntimeError(
                            "INTERNAL ERROR: Cannot resize dataset when pre-computing shapes."
                        )

                    if isinstance(value, Tensor) and value.is_cuda:  # type: ignore
                        value = value.cpu()  # type: ignore

                    # If the value is a sequence but not an array or tensor
                    if hdf_dtype in ("i", "f") and not isinstance(
                        value, (Tensor, np.ndarray)
                    ):
                        value = np.array(value)

                    # Note: "dset_audios[slices]" is a generic version of "dset_audios[i, :shape_0, :shape_1]"
                    slices = (i,) + tuple(slice(shape_i) for shape_i in shape)
                    try:
                        hdf_dset[slices] = value
                    except TypeError as err:
                        pylog.error(
                            f"Cannot set data {value} into {hdf_dset[slices].shape} ({attr_name=}, {i=}, {slices=})"
                        )
                        raise err

                    # Store original shape if needed
                    shape_name = f"{attr_name}{shape_suffix}"
                    if shape_name in hdf_dsets.keys():
                        hdf_shapes_dset = hdf_dsets[shape_name]
                        hdf_shapes_dset[i] = shape

                    global_hash_value += _checksum(value)

                i += 1

        # note: HDF cannot save too large int values with too many bits
        global_hash_value = global_hash_value % (2**31)

        attributes = {
            "creation_date": creation_date,
            "source_dataset": dataset.__class__.__name__,
            "length": len(dataset),
            "metadata": str(metadata),
            "encoding": HDF_ENCODING,
            "info": json.dumps(info),
            "global_hash_value": global_hash_value,
            "item_type": item_type,
            "added_columns": added_columns,
            "shape_suffix": shape_suffix,
        }
        if verbose >= 2:
            dumped_attributes = json.dumps(attributes, indent="\t")
            pylog.debug(f"Saving attributes in HDF file:\n{dumped_attributes}")

        for attr_name, attr_val in attributes.items():
            try:
                hdf_file.attrs[attr_name] = attr_val
            except TypeError as err:
                pylog.error(
                    f"Cannot store attribute {attr_name=} with value {attr_val=} in HDF."
                )
                raise err

    if verbose >= 2:
        pylog.debug(f"Data into has been packed into HDF file '{hdf_fpath}'.")

    hdf_dataset = HDFDataset(hdf_fpath, open_hdf=True, return_added_columns=False)
    return hdf_dataset


class Compose:
    def __init__(self, *fns: Callable) -> None:
        super().__init__()
        self.fns = fns

    def __call__(self, x: Any) -> Any:
        for fn in self.fns:
            x = fn(x)
        return x


def _checksum(
    value: Any,
) -> int:
    if isinstance(value, bytes):
        return zlib.adler32(value)
    elif isinstance(value, (np.ndarray, Tensor)):
        return int(value.sum().item())
    elif isinstance(value, (int, float)):
        return int(value)
    elif isinstance(value, str):
        return _checksum(value.encode())
    elif isinstance(value, (list, tuple)):
        return sum(map(_checksum, value))
    else:
        raise TypeError(f"Invalid argument type {value.__class__.__name__}.")


def _get_shape_and_dtype(
    value: Union[int, float, str, Tensor, list]
) -> Tuple[Tuple[int, ...], _HDFDType]:
    """Returns the shape and the hdf_dtype for an input."""
    if isinstance(value, int):
        shape = ()
        hdf_dtype = "i"

    elif isinstance(value, float):
        shape = ()
        hdf_dtype = "f"

    elif isinstance(value, str):
        shape = ()
        hdf_dtype = HDF_STRING_DTYPE

    elif isinstance(value, Tensor):
        shape = tuple(value.shape)
        if value.is_floating_point():
            hdf_dtype = "f"
        else:
            hdf_dtype = "i"

    elif isinstance(value, np.ndarray):
        shape = tuple(value.shape)
        hdf_dtype = value.dtype.kind
        if hdf_dtype == "u":
            hdf_dtype = "i"

    elif isinstance(value, (list, tuple)):
        if len(value) == 0:
            shape = (0,)
            hdf_dtype = HDF_VOID_DTYPE
        else:
            sub_shapes_and_dtypes = list(map(_get_shape_and_dtype, value))
            sub_shapes = [shape for shape, _ in sub_shapes_and_dtypes]
            sub_dtypes = [dtype for _, dtype in sub_shapes_and_dtypes]
            sub_dims = list(map(len, sub_shapes))

            if not all_eq(sub_dims):
                raise TypeError(
                    f"Unsupported list of heterogeneous shapes lengths. (found {sub_dims=})"
                )
            if not all_eq(sub_dtypes):
                raise TypeError(
                    f"Unsupported list of heterogeneous types. (found {sub_dtypes=})"
                )
            # Check for avoid ragged array like [["a", "b"], ["c"], ["d", "e"]]
            if not all_eq(sub_shapes):
                raise TypeError(
                    f"Unsupported list of heterogeneous shapes. (found {sub_shapes=} for {value=})"
                )

            max_subshape = tuple(
                max(shape[i] for shape in sub_shapes) for i in range(len(sub_shapes[0]))
            )
            shape = (len(value),) + max_subshape
            hdf_dtype = sub_dtypes[0]
    else:
        raise TypeError(
            f"Unsupported type {value.__class__.__name__} in function get_shape_and_dtype."
        )

    return shape, hdf_dtype



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.others import can_be_converted_to_tensor


class TestOthers(TestCase):
    def test_can_be_converted_to_tensor_1(self) -> None:
        lst = [[1, 0, 0], [2, 3, 4]]
        self.assertTrue(can_be_converted_to_tensor(lst))

    def test_can_be_converted_to_tensor_2(self) -> None:
        lst = [[1, 0, 0], [2, 3]]
        self.assertFalse(can_be_converted_to_tensor(lst))

    def test_can_be_converted_to_tensor_3(self) -> None:
        lst = [[[True], [False]], [[False], [True]]]
        self.assertTrue(can_be_converted_to_tensor(lst))

    def test_can_be_converted_to_tensor_4(self) -> None:
        lst = [[[]], [[]]]
        self.assertTrue(can_be_converted_to_tensor(lst))

    def test_can_be_converted_to_tensor_5(self) -> None:
        lst = [torch.rand(10), torch.rand(10)]
        self.assertFalse(can_be_converted_to_tensor(lst))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from torch import Tensor

from torchoutil.utils.ckpt import ModelCheckpointRegister


class TestRegister(TestCase):
    def test_cnext_register(self) -> None:
        register = ModelCheckpointRegister(
            infos={
                "cnext_bl": {
                    "architecture": "ConvNeXt",
                    "url": "https://zenodo.org/record/8020843/files/convnext_tiny_465mAP_BL_AC_70kit.pth?download=1",
                    "hash": "0688ae503f5893be0b6b71cb92f8b428",
                    "fname": "convnext_tiny_465mAP_BL_AC_70kit.pth",
                },
            },
            state_dict_key="model",
        )

        model_name = "cnext_bl"
        register.download_ckpt(model_name, force=True)
        state_dict = register.load_state_dict(model_name, offline=True, device="cpu")

        assert isinstance(state_dict, dict)
        assert all(isinstance(k, str) for k in state_dict.keys())
        assert all(isinstance(v, Tensor) for v in state_dict.values())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.crop import crop_dims


class TestReadme(TestCase):
    def test_crop_dims_example_1(self) -> None:
        x = torch.zeros(10, 10, 10)
        result = crop_dims(x, [1, 2, 3], dims="auto")
        expected = torch.zeros(1, 2, 3)
        self.assertTrue(torch.equal(result, expected))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union

import torch
from torch import Tensor, nn

from torchoutil.utils.packaging import _NUMPY_AVAILABLE


if _NUMPY_AVAILABLE:
    import numpy as np

    from torchoutil.nn.functional.numpy import to_numpy, from_numpy

    class ToNumpy(nn.Module):
        def __init__(self, dtype: Union[str, np.dtype, None] = None) -> None:
            super().__init__()
            self.dtype = dtype

        def forward(self, x: Union[Tensor, np.ndarray, list]) -> np.ndarray:
            return to_numpy(x, self.dtype)

    class FromNumpy(nn.Module):
        def __init__(
            self,
            dtype: Union[torch.dtype, None] = None,
            device: Union[str, torch.device, None] = None,
        ) -> None:
            super().__init__()
            self.device = device
            self.dtype = dtype

        def forward(self, x: np.ndarray) -> Tensor:
            return from_numpy(x, self.dtype, self.device)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, Iterable, Mapping, Union

from typing_extensions import TypeGuard


def is_dict_str(x: Any) -> TypeGuard[Dict[str, Any]]:
    return isinstance(x, dict) and all(isinstance(key, str) for key in x.keys())


def is_iterable_int(x: Any) -> TypeGuard[Iterable[int]]:
    return isinstance(x, Iterable) and all(isinstance(xi, int) for xi in x)


def is_iterable_str(x: Any) -> TypeGuard[Iterable[str]]:
    return not isinstance(x, str) and (
        isinstance(x, Iterable) and all(isinstance(xi, str) for xi in x)
    )


def is_iterable_bytes_list(x: Any) -> TypeGuard[Iterable[Union[bytes, list]]]:
    return isinstance(x, Iterable) and all(isinstance(xi, (bytes, list)) for xi in x)


def is_iterable_iterable_int(x: Any) -> TypeGuard[Iterable[Iterable[int]]]:
    return (
        isinstance(x, Iterable)
        and all(isinstance(xi, Iterable) for xi in x)
        and all(isinstance(xij, int) for xi in x for xij in xi)
    )


def is_mapping_str_any(x: Any) -> TypeGuard[Mapping[str, Any]]:
    return isinstance(x, Mapping) and all(isinstance(key, str) for key in x.keys())



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from torchoutil.utils.packaging import _H5PY_AVAILABLE

if not _H5PY_AVAILABLE:
    raise ImportError(
        "Optional dependancy 'h5py' is not installed. Please install it using 'pip install torchoutil[extras]'"
    )

del _H5PY_AVAILABLE


from torchoutil.utils.hdf.dataset import HDFDataset
from torchoutil.utils.hdf.pack import pack_to_hdf



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Type used for backward compatibility."""

from typing import NamedTuple

from torch import Tensor, __version__
from torch.torch_version import TorchVersion

if __version__ < TorchVersion("2.0.0"):
    namedtuple_values_indices = NamedTuple(
        "namedtuple_values_indices", [("values", Tensor), ("indices", Tensor)]
    )
    min = namedtuple_values_indices
    max = namedtuple_values_indices

else:
    from torch.return_types import max, min  # type: ignore # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import glob
import logging
import os.path as osp
from pathlib import Path
from typing import Any, Dict, Iterable, List, TypedDict, Union

from typing_extensions import NotRequired

from torchoutil.utils.packaging import _TENSORBOARD_AVAILABLE

if not _TENSORBOARD_AVAILABLE:
    raise ImportError(
        "Optional dependancy 'tensorboard' is not installed. Please install it using 'pip install torchoutil[extras]'"
    )

from tensorboard.backend.event_processing.event_file_loader import (  # type: ignore
    EventFileLoader,
)

logger = logging.getLogger(__name__)


_EVENT_FILE_PREFIX = "events.out.tfevents."
_DT_FLOAT = 1
_DT_STRING = 7
_DTYPES = (_DT_FLOAT, _DT_STRING)


class TensorboardEvent(TypedDict):
    wall_time: float
    step: int
    tag: str
    dtype: str
    value: Union[str, float]
    string_val: NotRequired[str]
    float_val: NotRequired[List[float]]


def load_event_file(
    fpath: Union[str, Path],
    cast_float_and_str: bool = True,
    ignore_underscore_tags: bool = True,
    verbose: int = 0,
) -> List[TensorboardEvent]:
    """
    Args:
        fpath: File path to a tensorboard event file.
        cast_float_and_str: Cast string to floats and store result in 'value' field. defaults to True.
        ignore_underscore_tags: Ignore event when tag starts with an underscore. defaults to True.
        verbose: Verbose level. Higher value means more log messages. defaults to 0.
    """
    if not osp.isfile(fpath):
        raise FileNotFoundError(f"Invalid argument {fpath=}. (not a file)")

    event_file_loader = EventFileLoader(fpath)
    raw_data = []

    for event in event_file_loader.Load():
        wall_time: float = event.wall_time  # type: ignore
        event_values: list = event.summary.value  # type: ignore
        step: int = event.step  # type: ignore

        for event_value in event_values:
            tag = event_value.tag
            dtype = event_value.tensor.dtype
            string_val = event_value.tensor.string_val
            float_val = event_value.tensor.float_val

            data_i = {
                "wall_time": wall_time,
                "step": step,
                "tag": tag,
                "dtype": dtype,
                "string_val": string_val,
                "float_val": float_val,
            }
            raw_data.append(data_i)

    data = []
    for data_i in raw_data:
        tag: str = data_i["tag"]
        dtype: Any = data_i["dtype"]

        if ignore_underscore_tags and tag.startswith("_"):
            if verbose >= 2:
                logger.debug(
                    f'Skip value with tag "{tag}" which begins by an underscore.'
                )
            continue

        if dtype == _DT_FLOAT:
            float_val: List[float] = data_i["float_val"]
            dtype = "float"

            if cast_float_and_str:
                assert len(float_val) == 1
                value = float_val[0]
                del data_i["string_val"]
                del data_i["float_val"]
            else:
                value = float_val

        elif dtype == _DT_STRING:
            string_val: str = data_i["string_val"]
            dtype = "str"

            if cast_float_and_str:
                value = string_val[3:-2]
                tag = tag.split("/")[0]
                del data_i["string_val"]
                del data_i["float_val"]
            else:
                value = string_val

        else:
            raise RuntimeError(f"Unknown value {dtype=}. (expected one of {_DTYPES})")

        data_i["tag"] = tag
        data_i["dtype"] = dtype
        data_i["value"] = value

        data.append(data_i)

    return data


def load_event_files(
    paths_or_patterns: Union[str, Path, Iterable[Union[str, Path]]],
    cast_float_and_str: bool = True,
    ignore_underscore_tags: bool = True,
    verbose: int = 0,
) -> Dict[str, List[TensorboardEvent]]:
    """
    Args:
        paths_or_patterns: Path or glob patterns to multiple files.
        cast_float_and_str: Cast string to floats and store result in 'value' field. defaults to True.
        ignore_underscore_tags: Ignore event when tag starts with an underscore. defaults to True.
        verbose: Verbose level. Higher value means more log messages. defaults to 0.
    """
    if isinstance(paths_or_patterns, (str, Path)):
        paths_or_patterns = [str(paths_or_patterns)]
    else:
        paths_or_patterns = [
            str(path_or_pattern) for path_or_pattern in paths_or_patterns
        ]

    paths = [
        path
        for path_or_pattern in paths_or_patterns
        for path in glob.iglob(path_or_pattern)
    ]
    all_events = {}
    for path in paths:
        events = load_event_file(
            path,
            cast_float_and_str,
            ignore_underscore_tags,
            verbose,
        )
        all_events[path] = events

    return all_events


def get_duration(
    fpath: Union[str, Path],
    verbose: int = 0,
) -> float:
    """Return time elapsed between first and last log in a tensorboard event file."""
    events = load_event_file(fpath, cast_float_and_str=True, verbose=verbose)
    wall_times = [event["wall_time"] for event in events]
    duration = max(wall_times) - min(wall_times)
    return duration



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, Iterable, List, Optional

import torch

from torchoutil.nn.functional.others import can_be_converted_to_tensor, can_be_stacked
from torchoutil.nn.functional.pad import pad_and_stack_rec
from torchoutil.utils.collections import filter_iterable, list_dict_to_dict_list


class CollateDict:
    """Collate class to handle a batch dict."""

    def __init__(self, error_on_missing_key: bool = True) -> None:
        super().__init__()
        self.error_on_missing_key = error_on_missing_key

    def __call__(self, batch_lst: List[Dict[str, Any]]) -> Dict[str, List[Any]]:
        result = list_dict_to_dict_list(
            batch_lst, error_on_missing_key=self.error_on_missing_key
        )
        return result


class AdvancedCollateDict:
    """Collate class to automatically convert to tensor, pad sequences and filter keys in a batch dict."""

    def __init__(
        self,
        pad_values: Optional[Dict[str, Any]] = None,
        include_keys: Optional[Iterable[str]] = None,
        exclude_keys: Optional[Iterable[str]] = None,
        error_on_missing_key: bool = True,
    ) -> None:
        """Collate list of dict into a dict of list WITH auto-padding for given keys."""
        if pad_values is None:
            pad_values = {}

        super().__init__()
        self.pad_values = pad_values
        self.include_keys = include_keys
        self.exclude_keys = exclude_keys
        self.error_on_missing_key = error_on_missing_key

    def __call__(self, batch_lst: List[Dict[str, Any]]) -> Dict[str, Any]:
        batch_dict = list_dict_to_dict_list(
            batch_lst, error_on_missing_key=self.error_on_missing_key
        )
        batch_keys = filter_iterable(
            batch_dict.keys(), self.include_keys, self.exclude_keys
        )
        batch_dict = {k: batch_dict[k] for k in batch_keys}

        result = {}

        for key, values in batch_dict.items():
            if key in self.pad_values:
                values = pad_and_stack_rec(values, self.pad_values[key])
            elif can_be_stacked(values):
                values = torch.stack(values)
            elif can_be_converted_to_tensor(values):
                values = torch.as_tensor(values)

            result[key] = values

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from torch import Tensor


def softmax_multidim(
    input: Tensor,
    dims: Union[Iterable[int], None] = (-1,),
) -> Tensor:
    """A multi-dimensional version of torch.softmax."""
    if dims is not None:
        dims = tuple(dims)

    x = input.exp()
    result = x / x.sum(dim=dims, keepdim=True)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from importlib.util import find_spec


def _package_is_available(package_name: str) -> bool:
    """Returns True if package is installed in the current python environment."""
    try:
        return find_spec(package_name) is not None
    except AttributeError:
        # Old support for Python <= 3.6
        return False
    except (ImportError, ModuleNotFoundError):
        # Python >= 3.7
        return False


_NUMPY_AVAILABLE = _package_is_available("numpy")
_TENSORBOARD_AVAILABLE = _package_is_available("tensorboard")
_H5PY_AVAILABLE = _package_is_available("h5py")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.repeat import repeat_interleave_nd


class TestRepeat(TestCase):
    def test_example_1(self) -> None:
        x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        result = repeat_interleave_nd(x, repeats=2, dim=0)
        expected = torch.as_tensor(
            [[0, 1, 2, 3], [0, 1, 2, 3], [4, 5, 6, 7], [4, 5, 6, 7]]
        )
        self.assertTrue(torch.equal(result, expected))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Generic, List, Mapping, TypeVar, Union

import torch
from torch import Tensor, nn

from torchoutil.nn.functional.multilabel import (
    indices_to_multihot,
    indices_to_names,
    multihot_to_indices,
    multihot_to_names,
    names_to_indices,
    names_to_multihot,
    probs_to_indices,
    probs_to_multihot,
    probs_to_names,
)
from torchoutil.utils.collections import dump_dict

T = TypeVar("T")


class IndicesToMultihot(nn.Module):
    def __init__(
        self,
        num_classes: int,
        *,
        device: Union[str, torch.device, None] = None,
        dtype: Union[torch.dtype, None] = torch.bool,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        indices: Union[List[List[int]], List[Tensor]],
    ) -> Tensor:
        multihot = indices_to_multihot(
            indices,
            self.num_classes,
            device=self.device,
            dtype=self.dtype,
        )
        return multihot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                num_classes=self.num_classes,
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_none=True,
        )


class IndicesToNames(nn.Module, Generic[T]):
    def __init__(
        self,
        idx_to_name: Mapping[int, T],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        indices: Union[List[List[int]], List[Tensor]],
    ) -> List[List[T]]:
        names = indices_to_names(indices, self.idx_to_name)
        return names


class MultihotToIndices(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(
        self,
        multihot: Tensor,
    ) -> List[List[int]]:
        names = multihot_to_indices(multihot)
        return names


class MultihotToNames(nn.Module, Generic[T]):
    def __init__(
        self,
        idx_to_name: Mapping[int, T],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        multihot: Tensor,
    ) -> List[List[T]]:
        names = multihot_to_names(multihot, self.idx_to_name)
        return names


class NamesToIndices(nn.Module, Generic[T]):
    def __init__(
        self,
        idx_to_name: Mapping[int, T],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        names: List[List[T]],
    ) -> List[List[int]]:
        indices = names_to_indices(names, self.idx_to_name)
        return indices


class NamesToMultihot(nn.Module, Generic[T]):
    def __init__(
        self,
        idx_to_name: Mapping[int, T],
        *,
        device: Union[str, torch.device, None] = None,
        dtype: Union[torch.dtype, None] = torch.bool,
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        names: List[List[T]],
    ) -> Tensor:
        multihot = names_to_multihot(
            names,
            self.idx_to_name,
            device=self.device,
            dtype=self.dtype,
        )
        return multihot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_none=True,
        )


class ProbsToIndices(nn.Module):
    def __init__(
        self,
        threshold: Union[float, Tensor],
    ) -> None:
        super().__init__()
        self.threshold = threshold

    def forward(
        self,
        probs: Tensor,
    ) -> List[List[int]]:
        indices = probs_to_indices(probs, self.threshold)
        return indices


class ProbsToMultihot(nn.Module):
    def __init__(
        self,
        threshold: Union[float, Tensor],
        *,
        device: Union[str, torch.device, None] = None,
        dtype: Union[torch.dtype, None] = torch.bool,
    ) -> None:
        super().__init__()
        self.threshold = threshold
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        probs: Tensor,
    ) -> Tensor:
        multihot = probs_to_multihot(
            probs,
            self.threshold,
            device=self.device,
            dtype=self.dtype,
        )
        return multihot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_none=True,
        )


class ProbsToNames(nn.Module, Generic[T]):
    def __init__(
        self,
        threshold: Union[float, Tensor],
        idx_to_name: Mapping[int, T],
    ) -> None:
        super().__init__()
        self.threshold = threshold
        self.idx_to_name = idx_to_name

    def forward(
        self,
        probs: Tensor,
    ) -> List[List[T]]:
        names = probs_to_names(probs, self.threshold, self.idx_to_name)
        return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Literal, Union

import torch
from torch import Generator, Tensor, nn

from torchoutil.nn.functional.pad import pad_and_stack_rec, pad_dim, pad_dims, PadAlign
from torchoutil.utils.collections import dump_dict


class PadDim(nn.Module):
    def __init__(
        self,
        target_length: int,
        align: PadAlign = "left",
        pad_value: float = 0.0,
        dim: int = -1,
        mode: str = "constant",
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.target_length = target_length
        self.align = align
        self.pad_value = pad_value
        self.dim = dim
        self.mode = mode
        self.generator = generator

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return pad_dim(
            x,
            self.target_length,
            self.align,  # type: ignore
            self.pad_value,
            self.dim,
            self.mode,
            self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_length=self.target_length,
                align=self.align,
                pad_value=self.pad_value,
                dim=self.dim,
                mode=self.mode,
            )
        )


class PadDims(nn.Module):
    def __init__(
        self,
        target_lengths: Iterable[int],
        aligns: Iterable[PadAlign] = ("left",),
        pad_value: float = 0.0,
        dims: Iterable[int] = (-1,),
        mode: str = "constant",
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.target_lengths = target_lengths
        self.aligns = aligns
        self.pad_value = pad_value
        self.dims = dims
        self.mode = mode
        self.generator = generator

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return pad_dims(
            x,
            self.target_lengths,
            self.aligns,
            self.pad_value,
            self.dims,
            self.mode,
            self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_lengths=self.target_lengths,
                aligns=self.aligns,
                pad_value=self.pad_value,
                dims=self.dims,
                mode=self.mode,
            )
        )


class PadAndStackRec(nn.Module):
    def __init__(
        self,
        pad_value: float,
        dtype: Union[None, torch.dtype] = None,
        device: Union[str, torch.device, None] = None,
    ) -> None:
        super().__init__()
        self.pad_value = pad_value
        self.dtype = dtype
        self.device = device

    def forward(
        self,
        sequence: Union[Tensor, int, float, tuple, list],
    ) -> Tensor:
        return pad_and_stack_rec(sequence, self.pad_value, self.dtype, self.device)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                pad_value=self.pad_value,
                dtype=self.dtype,
                device=self.device,
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch
from torch import Tensor

from torchoutil.nn.modules.multilabel import (
    IndicesToMultihot,
    IndicesToNames,
    MultihotToIndices,
    MultihotToNames,
    NamesToIndices,
    NamesToMultihot,
)
from torchoutil.nn.modules.typed import TSequential


class TestMultilabel(TestCase):
    def test_modules(self) -> None:
        num_samples = int(torch.randint(0, 20, ()).item())
        num_classes = int(torch.randint(1, 20, ()).item())
        threshold = torch.rand(())
        idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))

        multihot = torch.rand(num_samples, num_classes).ge(threshold)

        # dummy pipeline to convert labels multiple times
        pipeline = TSequential[Tensor, Tensor](
            MultihotToIndices(),
            IndicesToNames(idx_to_name),
            NamesToMultihot(idx_to_name),
            MultihotToNames(idx_to_name),
            NamesToIndices(idx_to_name),
            IndicesToMultihot(num_classes),
        )
        result = pipeline(multihot)
        self.assertTrue(torch.equal(multihot, result))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Dict, Tuple, TypeVar

import h5py

T = TypeVar("T")


# Force this encoding
HDF_ENCODING = "utf-8"
# Key suffix to store tensor shapes (because they are padded in hdf file)
SHAPE_SUFFIX = "_shape"
# Type for strings
HDF_STRING_DTYPE = h5py.string_dtype(HDF_ENCODING, None)
# Type for empty lists
HDF_VOID_DTYPE = h5py.opaque_dtype("V1")


def _tuple_to_dict(x: Tuple[T, ...]) -> Dict[str, T]:
    return dict(zip(map(str, range(len(x))), x))


def _dict_to_tuple(x: Dict[str, T]) -> Tuple[T, ...]:
    return tuple(x.values())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Sized,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from torch import Tensor, nn
from typing_extensions import TypeGuard

from torchoutil.nn.functional.get import get_device


T = TypeVar("T")
U = TypeVar("U")


def find(
    x: Tensor,
    value: Any,
    default: Union[None, Tensor, int, float] = None,
    dim: int = -1,
) -> Tensor:
    """Return the index of the first occurrence of value in a tensor."""
    if x.ndim == 0:
        raise ValueError(
            f"Function find does not supports 0-d tensors. (found {x.ndim=} == 0)"
        )
    mask = x.eq(value)
    contains_eos = mask.any(dim=dim)
    indices_eos = mask.int().argmax(dim=dim)

    if default is None:
        if not contains_eos.all():
            raise RuntimeError(f"Cannot find {value=} in tensor.")
        return indices_eos
    else:
        output = torch.where(contains_eos, indices_eos, default)
        return output


@overload
def move_to_rec(
    x: Mapping[T, U],
    predicate: Optional[Callable[[Any], bool]] = None,
    **kwargs,
) -> Dict[T, U]:
    ...


@overload
def move_to_rec(
    x: T,
    predicate: Optional[Callable[[Any], bool]] = None,
    **kwargs,
) -> T:
    ...


def move_to_rec(
    x: Any,
    predicate: Optional[Callable[[Any], bool]] = None,
    **kwargs,
) -> Any:
    """Move all modules and tensors recursively to a specific dtype or device."""
    if "device" in kwargs:
        kwargs["device"] = get_device(kwargs["device"])

    if isinstance(x, (Tensor, nn.Module)):
        if predicate is None or predicate(x):
            return x.to(**kwargs)
        else:
            return x
    elif isinstance(x, (str, float, int, bool, complex)):
        return x
    elif isinstance(x, Mapping):
        return {k: move_to_rec(v, predicate=predicate, **kwargs) for k, v in x.items()}
    elif isinstance(x, Iterable):
        generator = (move_to_rec(xi, predicate=predicate, **kwargs) for xi in x)
        if isinstance(x, Generator):
            return generator
        elif isinstance(x, tuple):
            return tuple(generator)
        else:
            return list(generator)
    else:
        return x


def is_scalar(x: Any) -> TypeGuard[Union[int, float, bool, complex, Tensor]]:
    return isinstance(x, (int, float, bool, complex)) or (
        isinstance(x, Tensor) and x.ndim == 0
    )


def can_be_stacked(
    tensors: Union[List[Any], Tuple[Any, ...]]
) -> TypeGuard[Union[List[Tensor], Tuple[Tensor, ...]]]:
    if len(tensors) == 0:
        return True
    if not all(isinstance(tensor, Tensor) for tensor in tensors):
        return False
    shape0 = tensors[0].shape
    result = all(tensor.shape == shape0 for tensor in tensors[1:])
    return result


def can_be_converted_to_tensor(x: Any) -> bool:
    if isinstance(x, (int, float, bool, complex)):
        return True
    elif isinstance(x, (List, Tuple)):
        return __can_be_converted_to_tensor_list_tuple(x)
    else:
        return False


def __can_be_converted_to_tensor_list_tuple(x: Union[List, Tuple]) -> bool:
    if len(x) == 0:
        return True

    valid_items = all(can_be_converted_to_tensor(xi) for xi in x)
    if not valid_items:
        return False

    if all(not isinstance(xi, Sized) for xi in x):
        return True
    elif all(isinstance(xi, Sized) for xi in x):
        len0 = len(x[0])
        return all(len(xi) == len0 for xi in x[1:])
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Generic, OrderedDict, TypeVar, overload

from torch import nn

InType = TypeVar("InType", covariant=False, contravariant=True)
OutType = TypeVar("OutType", covariant=True, contravariant=False)
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")


class TModule(Generic[InType, OutType], nn.Module):
    """Typed version of torch.nn.Module. Can specify an input and output type."""

    def __call__(self, *args: InType, **kwargs: InType) -> OutType:
        return super().__call__(*args, **kwargs)

    def forward(self, *args: InType, **kwargs: InType) -> OutType:
        return super().forward(*args, **kwargs)

    @overload
    def compose(self, other: "TModule[Any, T1]") -> "TSequential[InType, T1]":
        ...

    @overload
    def compose(self, other: nn.Module) -> "TSequential[InType, Any]":
        ...

    def compose(self, other) -> "TSequential[InType, Any]":
        return TSequential(self, other)

    def duplicate(self, num: int) -> "TSequential[InType, OutType]":
        duplicated = [self] * num
        return TSequential(*duplicated)


class TSequential(Generic[InType, OutType], TModule[InType, OutType], nn.Sequential):
    """Typed version of torch.nn.Sequential, designed to work with torchoutil.nn.TModules."""

    @overload
    def __init__(
        self,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, T1],
        arg1: TModule[T1, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, T1],
        arg1: TModule[T1, T2],
        arg2: TModule[T2, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, T1],
        arg1: TModule[T1, T2],
        arg2: TModule[T2, T3],
        arg3: TModule[T3, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, T1],
        arg1: TModule[T1, T2],
        arg2: TModule[T2, T3],
        arg3: TModule[T3, T4],
        arg4: TModule[T4, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, T1],
        arg1: TModule[T1, T2],
        arg2: TModule[T2, T3],
        arg3: TModule[T3, T4],
        arg4: TModule[T4, T5],
        arg5: TModule[T5, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, T1],
        arg1: TModule[T1, T2],
        arg2: TModule[T2, T3],
        arg3: TModule[T3, T4],
        arg4: TModule[T4, T5],
        arg5: TModule[T5, T6],
        arg6: TModule[T6, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, TModule[InType, OutType]]",
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, nn.Module]",
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *args: nn.Module,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        ...

    def __init__(
        self,
        *args,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:  # type: ignore
        nn.Sequential.__init__(self, *args)
        TModule.__init__(self)

        self.__unpack_tuple = unpack_tuple
        self.__unpack_dict = unpack_dict

    def __call__(self, x: InType) -> OutType:
        return nn.Sequential.__call__(self, x)

    def forward(self, x: InType) -> OutType:
        for module in self:
            if self.__unpack_tuple and isinstance(x, tuple):
                x = module(*x)
            elif self.__unpack_dict and isinstance(x, dict):
                x = module(**x)
            else:
                x = module(x)
        return x  # type: ignore


def __test_typing() -> None:
    import torch
    from torch import Tensor

    class LayerA(TModule[Tensor, Tensor]):
        def forward(self, x: Tensor) -> Tensor:
            return x * x

    class LayerB(TModule[Tensor, int]):
        def forward(self, x: Tensor) -> int:
            return int(x.sum().item())

    class LayerC(TModule[int, Tensor]):
        def forward(self, x: int) -> Tensor:
            return torch.as_tensor(x)

    x = torch.rand(10)
    xa = LayerA()(x)
    xb = LayerB()(x)

    seq = TSequential(LayerA(), LayerA(), LayerB())
    xab = seq(x)

    seq = LayerA().compose(LayerA()).compose(LayerB())
    xab = seq(x)

    seq = LayerC().compose(LayerA())
    xc = seq(2)

    assert isinstance(xa, Tensor)
    assert isinstance(xb, int)
    assert isinstance(xab, int)
    assert isinstance(xc, Tensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, Iterable, List, Literal, Sized, Tuple, Union

import torch
from torch import Generator, Size, Tensor
from torch.nn import functional as F
from torch.types import Number

from torchoutil.nn.functional.get import get_device
from torchoutil.nn.functional.others import can_be_stacked, is_scalar

PAD_ALIGNS = ("left", "right", "center", "random")
PadAlign = Literal["left", "right", "center", "random"]


def pad_dim(
    x: Tensor,
    target_length: int,
    align: PadAlign = "left",
    pad_value: float = 0.0,
    dim: int = -1,
    mode: str = "constant",
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Generic function for pad a single dimension."""
    return pad_dims(x, [target_length], [align], pad_value, [dim], mode, generator)


def pad_dims(
    x: Tensor,
    target_lengths: Iterable[int],
    aligns: Iterable[PadAlign] = ("left",),
    pad_value: float = 0.0,
    dims: Iterable[int] = (-1,),
    mode: str = "constant",
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Generic function to pad multiple dimensions."""
    if isinstance(generator, int):
        generator = Generator().manual_seed(generator)

    target_lengths = list(target_lengths)
    aligns = list(aligns)
    dims = list(dims)

    if len(dims) == 0:
        raise ValueError(
            f"Invalid argument {dims=}. (cannot use an empty list of dimensions)"
        )

    if len(target_lengths) != len(dims):
        raise ValueError(
            f"Invalid number of targets lengths ({len(target_lengths)}) with the number of dimensions ({len(dims)})."
        )

    if len(aligns) != len(dims):
        raise ValueError(
            f"Invalid number of aligns ({len(aligns)}) with the number of dimensions ({len(dims)})."
        )

    pad_seq = __generate_pad_seq(x.shape, target_lengths, dims, aligns, generator)
    x = F.pad(x, pad_seq, mode=mode, value=pad_value)
    return x


def pad_and_stack_rec(
    sequence: Union[Tensor, int, float, tuple, list],
    pad_value: Number,
    dtype: Union[None, torch.dtype] = None,
    device: Union[str, torch.device, None] = None,
) -> Tensor:
    """Recursive version of torch.nn.utils.rnn.pad_sequence, with padding of Tensors.

    Args:
        sequence: The sequence to pad. Must be convertable to tensor by having the correct number of dims in all sublists.
        pad_value: The pad value used.
        dtype: The dtype of the output Tensor. defaults to None.
        device: The device of the output Tensor. defaults to None.

    Example 1::
    -----------
        >>> sequence = [[1, 2], [3], [], [4, 5]]
        >>> output = pad_sequence_rec(sequence, 0)
        tensor([[1, 2], [3, 0], [0, 0], [4, 5]])

    Example 2::
    -----------
        >>> invalid_sequence = [[1, 2, 3], 3]
        >>> output = pad_sequence_rec(invalid_sequence, 0)
        ValueError : Cannot pad sequence of tensors of differents number of dims.

    """
    device = get_device(device)

    if isinstance(sequence, Tensor):
        return sequence.to(dtype=dtype, device=device)

    elif is_scalar(sequence) or (isinstance(sequence, Sized) and len(sequence) == 0):
        return torch.as_tensor(sequence, dtype=dtype, device=device)  # type: ignore

    elif isinstance(sequence, (list, tuple)):
        sequence = [
            pad_and_stack_rec(elt, pad_value, dtype, device) for elt in sequence
        ]
        if can_be_stacked(sequence):
            return torch.stack(sequence)

        shapes = [elt.shape for elt in sequence]
        shape0 = shapes[0]

        if not all(len(shape) == len(shape0) for shape in shapes):
            raise ValueError(
                f"Cannot pad sequence of tensors of differents number of dims. (with {shapes=})"
            )

        max_lens = [max(shape[i] for shape in shapes) for i in range(len(shape0))]
        sequence = [
            pad_dims(
                xi,
                target_lengths=max_lens,
                pad_value=pad_value,
                aligns=["left"] * xi.ndim,
                dims=range(xi.ndim),
            )
            for xi in sequence
        ]
        result = torch.stack(sequence)
        return result

    else:
        raise TypeError(
            f"Invalid type {type(sequence)}. (expected Tensor, int, float, list or tuple)"
        )


def __generate_pad_seq(
    x_shape: Size,
    target_lengths: List[int],
    dims: List[int],
    aligns: List[PadAlign],
    generator: Union[None, Generator],
) -> List[int]:
    pad_seq = [0 for _ in range(len(x_shape) * 2)]
    for target_length, dim, align in zip(target_lengths, dims, aligns):
        missing = max(target_length - x_shape[dim], 0)

        if align == "left":
            missing_left = 0
            missing_right = missing
        elif align == "right":
            missing_left = missing
            missing_right = 0
        elif align == "center":
            missing_left = missing // 2 + missing % 2
            missing_right = missing // 2
        elif align == "random":
            missing_left = int(
                torch.randint(
                    low=0, high=missing + 1, size=(), generator=generator
                ).item()
            )
            missing_right = missing - missing_left
        else:
            raise ValueError(
                f"Invalid argument {align=}. (expected one of {PAD_ALIGNS})"
            )

        # Note: pad_seq : [pad_left_dim_-1, pad_right_dim_-1, pad_left_dim_-2, pad_right_dim_-2, ...)
        idx = len(x_shape) - (dim % len(x_shape)) - 1
        assert pad_seq[idx * 2] == 0 and pad_seq[idx * 2 + 1] == 0
        pad_seq[idx * 2] = missing_left
        pad_seq[idx * 2 + 1] = missing_right

    return pad_seq


def cat_padded_batch(
    x1: Tensor,
    x1_lens: Tensor,
    x2: Tensor,
    x2_lens: Tensor,
    seq_dim: int = -1,
    batch_dim: int = 0,
) -> Tuple[Tensor, Tensor]:
    """Concatenate padded batched of sequences.

    Args:
        x1: First batch with D dims of shape (batch_size, ..., N1, ...)
        x1_lens: First lengths of each element in sequence dim of shape (batch_size,).
        x2: Second batch with D dims of shape (batch_size, ..., N2, ...)
            The shape must be the same than x1 unless for the dimension N2.
        x2_lens: Second lengths of each element in sequence dim of shape (batch_size,).
        seq_dim: Dimension index of sequence. defaults to -1.
        batch_dim: Batch dimension index. defaults to 0.
    """
    _check_cat_padded_batch(x1, x1_lens, x2, x2_lens, seq_dim, batch_dim)
    x12_lens = x1_lens + x2_lens
    sum_size_12 = x1.shape[seq_dim] + x2.shape[seq_dim]

    x12 = pad_dim(x1, sum_size_12, dim=seq_dim)
    kwd: Dict[str, Any] = dict(device=x1.device, dtype=torch.long)
    indices = torch.arange(x2_lens.max().item(), **kwd)

    unsq_x1_lens = x1_lens
    ndim = x1.ndim
    for i in range(ndim):
        if i != (seq_dim % ndim):
            indices = indices.unsqueeze(dim=i)
        if i != (batch_dim % ndim):
            unsq_x1_lens = unsq_x1_lens.unsqueeze(dim=i)

    expand_size = list(x2.shape)
    expand_size[seq_dim] = -1
    indices = indices.expand(*expand_size)
    indices = indices + unsq_x1_lens
    x12.scatter_(seq_dim, indices, x2)

    max_size_12 = int(x12_lens.max().item())
    if max_size_12 < sum_size_12:
        slices = [slice(None) for _ in range(ndim)]
        slices[seq_dim] = slice(max_size_12)
        x12 = x12[slices]

    return x12, x12_lens


def _check_cat_padded_batch(
    x1: Tensor,
    x1_lens: Tensor,
    x2: Tensor,
    x2_lens: Tensor,
    seq_dim: int,
    batch_dim: int,
) -> None:
    if x1.ndim != x2.ndim:
        raise ValueError(f"Invalid arguments ndims. (found {x1.ndim=} != {x2.ndim=})")
    if x1.ndim < 2:
        raise ValueError(f"Invalid arguments ndims. (found {x1.ndim=} < 2)")

    batch_size = x1.shape[batch_dim]
    if not (x1_lens.shape == x2_lens.shape == Size((batch_size,))):
        raise ValueError(
            f"Invalid arguments shape. (with {x1_lens.shape=} and {x2_lens.shape=})"
        )

    x1_shape = torch.as_tensor(x1.shape)
    x2_shape = torch.as_tensor(x2.shape)
    eq_mask = x1_shape.eq(x2_shape)
    eq_mask[seq_dim] = True
    if not eq_mask.all():
        raise ValueError(f"Invalid arguments shape. (with {x1.shape=} and {x2.shape=})")
